import pytest


class TestPuppet:
    @pytest.mark.complete("puppet ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("puppet agent --")
    def test_2(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(pre_cmds=("PATH=$PWD/shared/bin:$PATH",))
class TestWol:
    @pytest.mark.complete("wol ")
    def test_1(self, completion):
        assert all(
            x in completion
            for x in "00:00:00:00:00:00 11:11:11:11:11:11 "
            "22:22:22:22:22:22 33:33:33:33:33:33".split()
        )

    @pytest.mark.complete("wol 00:")
    def test_2(self, completion):
        assert any(x.endswith("00:00:00:00:00") for x in completion)

    @pytest.mark.complete("wol -", require_cmd=True)
    def test_3(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(cmd="apt-get")
class TestAptGet:
    @pytest.mark.complete("apt-get ")
    def test_1(self, completion):
        assert all(x in completion for x in "install update".split())

    @pytest.mark.complete("apt-get install ./", cwd="dpkg")
    def test_2(self, completion):
        assert completion == "bash-completion-test-subject.deb"

    @pytest.mark.complete("apt-get build-dep ")
    def test_build_dep_dirs(self, completion):
        assert "dpkg/" in completion



import pytest


class TestXvnc4viewer:
    @pytest.mark.complete("xvnc4viewer -")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("xvnc4viewer -PreferredEncoding ")
    def test_2(self, completion):
        assert completion == "hextile raw zrle".split()

    @pytest.mark.complete("xvnc4viewer --preferredencoding ")
    def test_3(self, completion):
        assert completion == "hextile raw zrle".split()



import pytest


@pytest.mark.bashcomp(cmd="lintian-info")
class TestLintianInfo:
    @pytest.mark.complete("lintian-info ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("lintian-info --", require_cmd=True)
    def test_2(self, completion):
        assert completion



import pytest


class TestSetquota:
    @pytest.mark.complete("setquota ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("setquota -", require_cmd=True)
    def test_2(self, completion):
        assert completion



# Based on work by Stephen Gildea, October 2010.

import pytest

from conftest import assert_bash_exec


@pytest.mark.bashcomp(cmd=None, ignore_env=r"^\+COMPREPLY=")
class TestUnitCompleteLongopt:
    @pytest.fixture(scope="class")
    def functions(self, request, bash):
        assert_bash_exec(bash, "_grephelp() { cat _longopt/grep--help.txt; }")
        assert_bash_exec(bash, "complete -F _comp_complete_longopt _grephelp")
        assert_bash_exec(bash, "_various() { cat _longopt/various.txt; }")
        assert_bash_exec(bash, "complete -F _comp_complete_longopt _various")

    @pytest.mark.complete("_grephelp --")
    def test_1(self, functions, completion):
        """First long option should be included"""
        assert completion
        assert all(
            x in completion for x in "--quiet --recursive --text".split()
        )

    @pytest.mark.complete("_grephelp -")
    def test_2(self, functions, completion):
        """Only long options should be included"""
        assert completion
        assert all(x.startswith("--") for x in completion)

    @pytest.mark.complete("_grephelp --")
    def test_3(self, functions, completion):
        """Should have both ones ending with a = and ones not"""
        assert completion
        assert any(x.endswith("=") for x in completion)
        assert any(not x.endswith("=") for x in completion)

    @pytest.mark.complete("_various --")
    def test_no_dashdashdash(self, functions, completion):
        assert all(not x.startswith("---") for x in completion)

    @pytest.mark.complete("_various --")
    def test_no_trailingdash(self, functions, completion):
        assert all(not x.endswith("-") for x in completion)

    @pytest.mark.complete("_various --")
    def test_underscore(self, functions, completion):
        assert "--foo_bar" in completion

    @pytest.mark.complete("_various --")
    def test_equals(self, functions, completion):
        assert "--foo=" in completion



import pytest


@pytest.mark.bashcomp(cmd="tsig-keygen")
class TestTsigKeygen:
    @pytest.mark.complete("tsig-keygen ")
    def test_basic(self, completion):
        assert not completion

    @pytest.mark.complete("tsig-keygen -", require_cmd=True)
    def test_options(self, completion):
        assert completion



import pytest


class TestSh:
    @pytest.mark.complete("sh -")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("sh +")
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("sh -o ")
    def test_3(self, completion):
        assert completion

    @pytest.mark.complete("sh -c ")
    def test_4(self, completion):
        assert not completion



import pytest


class TestCreatedb:
    # --help can fail due to missing package dependencies, e.g. on Ubuntu 14
    @pytest.mark.complete(
        "createdb -", require_cmd=True, xfail="! createdb --help &>/dev/null"
    )
    def test_1(self, completion):
        assert completion



import pytest

from conftest import assert_bash_exec, bash_env_saved


@pytest.mark.bashcomp(cmd=None, ignore_env=r"^[+-](cur|COMPREPLY)=")
class TestUnitExpand:
    def test_1(self, bash):
        assert_bash_exec(bash, "_comp_expand >/dev/null")

    def test_2(self, bash):
        """Test environment non-pollution, detected at teardown."""
        assert_bash_exec(bash, "foo() { _comp_expand; }; foo; unset -f foo")

    def test_user_home_compreply(self, bash, user_home):
        user, home = user_home
        output = assert_bash_exec(
            bash,
            r'cur="~%s"; _comp_expand; printf "%%s\n" "$COMPREPLY"' % user,
            want_output=True,
        )
        assert output.strip() == home

    def test_user_home_compreply_failglob(self, bash, user_home):
        user, home = user_home
        with bash_env_saved(bash) as bash_env:
            bash_env.shopt("failglob", True)
            output = assert_bash_exec(
                bash,
                r'cur="~%s"; _comp_expand; printf "%%s\n" "$COMPREPLY"' % user,
                want_output=True,
            )
        assert output.strip() == home

    def test_user_home_cur(self, bash, user_home):
        user, home = user_home
        output = assert_bash_exec(
            bash,
            r'cur="~%s/a"; _comp_expand; printf "%%s\n" "$cur"' % user,
            want_output=True,
        )
        assert output.strip() == "%s/a" % home



import os

import pytest


class TestChsh:
    @pytest.mark.complete("chsh ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("chsh -s ")
    def test_shells(self, completion):
        if os.path.exists("/etc/shells"):
            assert completion
        else:
            assert not completion

    @pytest.mark.complete("chsh -", require_cmd=True)
    def test_3(self, completion):
        assert completion

    @pytest.mark.complete("chsh --root shells -s ")
    def test_chroot_shells(self, completion):
        assert completion == "/bash/completion/canary"



import os

import pytest


class TestTime:
    @pytest.mark.complete("time _comp_delimite", cwd="shared/empty_dir")
    def test_command(self, completion):
        """
        Test completion of commands.

        We use a function of ours as the test subject, as that's guaranteed
        to be available, and do not rely on anything in particular in $PATH.
        """
        assert completion == "d" or "_comp_delimited" in completion

    @pytest.mark.complete("time -p find -typ")
    def test_2(self, completion):
        assert completion  # find's options

    @pytest.mark.complete("time shared/bin/")
    def test_3(self, completion):
        execs = sorted(
            x
            for x in os.listdir("shared/bin")
            if os.path.isfile("shared/bin/%s" % x)
            and os.access("shared/bin/%s" % x, os.X_OK)
        )
        assert completion == execs



import pytest


class TestLvextend:
    @pytest.mark.complete(
        "lvextend --", require_cmd=True, xfail="! lvextend --help &>/dev/null"
    )
    def test_1(self, completion):
        assert completion



import pytest


class TestMd5sum:
    @pytest.mark.complete("md5sum ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("md5sum --", require_longopt=True)
    def test_options(self, completion):
        assert completion



from itertools import chain

import pytest

from conftest import assert_bash_exec, bash_env_saved


@pytest.mark.bashcomp(
    cmd=None,
    ignore_env="^[+-](COMPREPLY|BASH_COMPLETION_KNOWN_HOSTS_WITH_HOSTFILE)=",
)
class TestUnitCompgenKnownHosts:
    @pytest.mark.parametrize(
        "prefix,colon_flag,hostfile",
        [("", "", True), ("", "", False), ("user@", "c", True)],
    )
    def test_basic(
        self, bash, hosts, avahi_hosts, prefix, colon_flag, hostfile
    ):
        expected = (
            "%s%s%s" % (prefix, x, ":" if colon_flag else "")
            for x in chain(
                hosts if hostfile else avahi_hosts,
                # fixtures/_known_hosts/config
                "gee hus jar #not-a-comment".split(),
                # fixtures/_known_hosts/known_hosts
                (
                    "doo",
                    "ike",
                    "jub",
                    "10.0.0.1",
                    "kyl",
                    "100.0.0.2",
                    "10.10.0.3",
                    "blah",
                    "fd00:xxxx:xxxx:xxxx:xxxx:xxxx:xxxx:5555",
                    "fe80::123:0xff:dead:beef%eth0",
                    "1111:2222:3333:4444:5555:6666:xxxx:abab",
                    "11xx:2222:3333:4444:5555:6666:xxxx:abab",
                    "::42",
                ),
            )
        )
        assert_bash_exec(
            bash,
            "unset -v BASH_COMPLETION_KNOWN_HOSTS_WITH_HOSTFILE"
            if hostfile
            else "BASH_COMPLETION_KNOWN_HOSTS_WITH_HOSTFILE=",
        )
        output = assert_bash_exec(
            bash,
            "_comp_compgen_known_hosts -a%sF _known_hosts/config '%s'; "
            r'printf "%%s\n" "${COMPREPLY[@]}"; unset -v COMPREPLY'
            % (colon_flag, prefix),
            want_output=True,
        )
        assert sorted(set(output.split())) == sorted(expected)

    @pytest.mark.parametrize(
        "family,result",
        (
            ("4", "127.0.0.1 localhost"),
            ("6", "::1 localhost"),
            ("46", "localhost"),
        ),
    )
    def test_ip_filtering(self, bash, family, result):
        assert_bash_exec(
            bash,
            "unset -v COMPREPLY BASH_COMPLETION_KNOWN_HOSTS_WITH_HOSTFILE",
        )
        output = assert_bash_exec(
            bash,
            "BASH_COMPLETION_KNOWN_HOSTS_WITH_HOSTFILE= "
            "_comp_compgen_known_hosts -%sF _known_hosts/localhost_config ''; "
            r'printf "%%s\n" "${COMPREPLY[@]}"' % family,
            want_output=True,
        )
        assert sorted(set(output.strip().split())) == sorted(result.split())

    def test_consecutive_spaces(self, bash, hosts):
        expected = hosts.copy()
        # fixtures/_known_hosts/spaced  conf
        expected.extend("gee hus #not-a-comment".split())
        # fixtures/_known_hosts/known_hosts2
        expected.extend("two two2 two3 two4".split())
        # fixtures/_known_hosts_/spaced  known_hosts
        expected.extend("doo ike".split())

        output = assert_bash_exec(
            bash,
            "unset -v COMPREPLY BASH_COMPLETION_KNOWN_HOSTS_WITH_HOSTFILE; "
            "_comp_compgen_known_hosts -aF '_known_hosts/spaced  conf' ''; "
            r'printf "%s\n" "${COMPREPLY[@]}"',
            want_output=True,
        )
        assert sorted(set(output.strip().split())) == sorted(expected)

    def test_files_starting_with_tilde(self, bash, hosts):
        expected = hosts.copy()
        # fixtures/_known_hosts/known_hosts2
        expected.extend("two two2 two3 two4".split())
        # fixtures/_known_hosts/known_hosts3
        expected.append("three")
        # fixtures/_known_hosts/known_hosts4
        expected.append("four")

        with bash_env_saved(bash) as bash_env:
            bash_env.write_variable("HOME", bash.cwd)
            output = assert_bash_exec(
                bash,
                "unset -v COMPREPLY BASH_COMPLETION_KNOWN_HOSTS_WITH_HOSTFILE;"
                " _comp_compgen_known_hosts -aF _known_hosts/config_tilde ''; "
                r'printf "%s\n" "${COMPREPLY[@]}"',
                want_output=True,
            )

        assert sorted(set(output.strip().split())) == sorted(expected)

    def test_included_configs(self, bash, hosts):
        expected = hosts.copy()
        # fixtures/_known_hosts/config_include_recursion
        expected.append("recursion")
        # fixtures/_known_hosts/.ssh/config_relative_path
        expected.append("relative_path")
        # fixtures/_known_hosts/.ssh/config_asterisk_*
        expected.extend("asterisk_1 asterisk_2".split())
        # fixtures/_known_hosts/.ssh/config_question_mark
        expected.append("question_mark")

        with bash_env_saved(bash) as bash_env:
            bash_env.write_variable("HOME", "%s/_known_hosts" % bash.cwd)
            output = assert_bash_exec(
                bash,
                "unset -v COMPREPLY BASH_COMPLETION_KNOWN_HOSTS_WITH_HOSTFILE;"
                " _comp_compgen_known_hosts -aF _known_hosts/config_include ''; "
                r'printf "%s\n" "${COMPREPLY[@]}"',
                want_output=True,
            )
        assert sorted(set(output.strip().split())) == sorted(expected)

    def test_no_globbing(self, bash):
        with bash_env_saved(bash) as bash_env:
            bash_env.write_variable("HOME", "%s/_known_hosts" % bash.cwd)
            bash_env.chdir("_known_hosts")
            output = assert_bash_exec(
                bash,
                "unset -v COMPREPLY BASH_COMPLETION_KNOWN_HOSTS_WITH_HOSTFILE;"
                " _comp_compgen_known_hosts -aF config ''; "
                r'printf "%s\n" "${COMPREPLY[@]}"',
                want_output=True,
            )
        completion = sorted(set(output.strip().split()))
        assert "gee" in completion
        assert "gee-filename-canary" not in completion



import pytest


class TestIsort:
    @pytest.mark.complete("isort ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("isort -", require_longopt=True)
    def test_2(self, completion):
        assert completion



import pytest

from conftest import (
    assert_bash_exec,
    assert_complete,
    find_unique_completion_pair,
)


class TestLs:
    @pytest.mark.complete(
        "ls --", require_cmd=True, xfail="! ls --help &>/dev/null"
    )
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("ls ~")
    def test_2(self, completion):
        assert completion

    def test_3(self, bash):
        """~part should complete to ~full<SPACE> if home dir does not exist."""
        res = (
            assert_bash_exec(
                bash,
                "for u in $(compgen -u); do "
                "eval test -d ~$u || echo $u; unset -v u; done",
                want_output=True,
            )
            .strip()
            .split()
        )
        part_full = find_unique_completion_pair(res)
        if not part_full:
            pytest.skip("No suitable test user found")
            return
        part, full = part_full
        completion = assert_complete(bash, "ls ~%s" % part)
        assert completion == full[len(part) :]
        assert completion.endswith(" ")



import pytest


@pytest.mark.bashcomp(cmd="ssh-keygen")
class TestSshKeygen:
    @pytest.mark.complete("ssh-keygen -", require_cmd=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete(
        "ssh-keygen -", require_cmd=True, shopt=dict(failglob=True)
    )
    def test_1_failglob(self, completion):
        assert completion

    @pytest.mark.complete(
        "ssh-keygen -", require_cmd=True, shopt=dict(nullglob=True)
    )
    def test_1_nullglob(self, completion):
        assert completion

    @pytest.mark.complete("ssh-keygen -s foo_key ssh-copy-id/.ssh/")
    def test_filedir_pub_at_end_of_s(self, completion):
        assert completion
        assert all(x.endswith(".pub") for x in completion)

    @pytest.mark.complete("ssh-keygen -s foo_key -n foo,")
    def test_usernames_for_n(self, completion):
        assert completion
        # TODO check that these are usernames

    @pytest.mark.complete("ssh-keygen -s foo_key -h -n foo,")
    def test_host_for_h_n(self, completion):
        assert completion
        # TODO check that these are hostnames

    @pytest.mark.complete("ssh-keygen -Y foo -n ")
    def test_n_with_Y(self, completion):
        assert not completion

    @pytest.mark.complete("ssh-keygen -r ")
    def test_r_without_Y(self, completion):
        assert not completion

    @pytest.mark.complete("ssh-keygen -Y foo -r ")
    def test_r_with_Y(self, completion):
        assert "ssh/" in completion

    @pytest.mark.complete("ssh-keygen -t ecdsa -b ")
    def test_ecdsa_b(self, completion):
        assert completion

    @pytest.mark.complete("ssh-keygen -t ecdsa-sk -b ")
    def test_ecdsa_sk_b(self, completion):
        assert not completion

    @pytest.mark.complete("ssh-keygen -O ")
    def test_bare_O(self, completion):
        assert not completion

    @pytest.mark.complete("ssh-keygen -s -O ")
    def test_s_O(self, completion):
        assert completion
        assert any(x.endswith("=") for x in completion)

    @pytest.mark.complete("ssh-keygen -s -O force-command=bas")
    def test_O_force_command(self, completion):
        assert completion
        assert not completion.startswith("force-command=")

    @pytest.mark.complete("ssh-keygen -O unknown=")
    def test_O_unknown(self, completion):
        assert not completion

    @pytest.mark.complete("ssh-keygen -t ed25519-sk -O application=")
    def test_O_application(self, completion):
        assert completion == "ssh:"

    @pytest.mark.complete("ssh-keygen -t ed25519-sk -O application=s")
    def test_O_application_s(self, completion):
        assert completion == "sh:"

    @pytest.mark.complete("ssh-keygen -t ed25519-sk -O application=ssh:")
    def test_O_application_ssh_colon(self, completion):
        assert not completion

    @pytest.mark.complete(
        "ssh-keygen -t ed25519-sk -O application=nonexistent"
    )
    def test_O_application_nonexistent(self, completion):
        assert not completion

    @pytest.mark.complete("ssh-keygen -r -O ")
    def test_r_O(self, completion):
        assert completion

    @pytest.mark.complete("ssh-keygen -Y -O ")
    def test_Y_O(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(
    pre_cmds=(
        # Make sure our own ./configure is in PATH
        "PATH=$PWD/../..:$PATH",
    )
)
class TestConfigure:
    @pytest.mark.complete("configure --", require_cmd=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("configure --prefix ")
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("configure --unknown-option-with-split=")
    def test_unknown_split_filedir_fallback(self, completion):
        assert "shared/" in completion

    @pytest.mark.complete("configure --unknown-option ")
    def test_unknown_filedir_fallback(self, completion):
        assert "shared/" in completion



import pytest


class Test2to3:
    @pytest.mark.complete("2to3 ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("2to3 -", require_longopt=True)
    def test_2(self, completion):
        assert completion



import pytest

from conftest import assert_bash_exec


@pytest.mark.bashcomp(
    ignore_env=r"^\+ANT_ARGS=",
    temp_cwd=True,
    pre_cmds=('cp "$SRCDIRABS"/fixtures/ant/*.xml .',),
)
class TestAnt:
    @pytest.fixture(scope="class")
    def has_complete_ant_cmd_pl(self, bash):
        output = assert_bash_exec(bash, "complete -p ant", want_output=True)
        return "complete-ant-cmd.pl" in output

    @pytest.mark.complete("ant -", require_cmd=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("ant ")
    def test_2(self, completion):
        assert completion == "bashcomp clean init realclean".split()

    @pytest.mark.complete("ant -f build-with-import.xml ")
    def test_3(self, completion, has_complete_ant_cmd_pl):
        if has_complete_ant_cmd_pl:
            # Some versions of complete-ant-cmd.pl add "import-project-name."
            # prefix to imported targets, just check that the ones we add
            # are there.
            assert all(
                x in completion
                for x in "build-with-import imported-build".split()
            )
        else:
            assert completion == "build-with-import imported-build".split()

    @pytest.mark.complete("ant ", env=dict(ANT_ARGS="'-f named-build.xml'"))
    def test_4(self, bash, completion, has_complete_ant_cmd_pl):
        if has_complete_ant_cmd_pl:
            # Some versions of complete-ant-cmd.pl don't treat ANT_ARGS right;
            # in those cases we get the correct completion produced by us
            # plus whatever complete-ant-cmd.pl was able to get from build.xml
            assert "named-build" in completion
        else:
            assert completion == "named-build"

    @pytest.mark.complete("ant -l ")
    def test_5(self, completion):
        assert completion



import pytest


class TestRfkill:
    @pytest.mark.complete("rfkill ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("rfkill -")
    def test_2(self, completion):
        assert completion



import pytest


class TestFeh:
    @pytest.mark.complete("feh ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete(
        "feh --lis",
        xfail="feh --help 2>&1 | command grep -qF 'man feh'",
        require_cmd=True,
    )
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("feh -S pix")
    def test_3(self, completion):
        assert completion == "els"

    @pytest.mark.complete("feh --zoom ma")
    def test_4(self, completion):
        assert completion == "x"

    @pytest.mark.complete("feh -g 640")
    def test_5(self, completion):
        assert completion == "0 1 2 3 4 5 6 7 8 9 x".split()

    @pytest.mark.complete("feh -g 640x48")
    def test_6(self, completion):
        assert completion == "0 1 2 3 4 5 6 7 8 9".split()



import pytest


@pytest.mark.bashcomp(pre_cmds=("export PERL5LIB=$PWD/perldoc",))
class TestPerldoc:
    @pytest.mark.complete("perldoc File::")
    def test_1(self, completion):
        assert "Path" in completion  # Assume File::Path always installed
        assert "fixtures/" not in completion  # Our fixtures/ dir
        assert not [x for x in completion if "File::File::" in x]

    @pytest.mark.complete(
        "perldoc -",
        require_cmd=True,
        skipif="! perldoc -V &>/dev/null",
    )
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("perldoc BashCompletion")
    def test_3(self, completion):
        assert completion == "BashCompletionDoc BashCompletionModule".split()



import pytest


class TestDir:
    @pytest.mark.complete("dir ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("dir -", require_cmd=True)
    def test_options(self, completion):
        assert completion



import pytest


class TestCrontab:
    @pytest.mark.complete("crontab ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("crontab -l -")
    def test_only_u_with_l(self, completion):
        assert completion == "u"

    @pytest.mark.complete("crontab -r -")
    def test_no_l_with_r(self, completion):
        assert completion
        assert "-l" not in completion



import pytest

from conftest import assert_complete, partialize


class TestSsh:
    @pytest.mark.complete("ssh -Fsp", cwd="ssh")
    def test_1(self, completion):
        assert completion == r"aced\ \ conf"

    @pytest.mark.complete("ssh -F config ls", cwd="ssh")
    def test_2(self, completion):
        """
        Should not complete commands when host is not specified.

        Test sanity assumes there are commands starting with `ls`.
        """
        assert completion == "_known_host"

    @pytest.mark.complete("ssh bash", cwd="ssh")
    def test_3(self, completion):
        """
        First arg should not complete with commands.

        Assumes there's no "bash" known host.
        """
        assert "bash" not in completion

    @pytest.mark.complete("ssh -vo AddressFamily=")
    def test_4(self, completion):
        assert completion

    @pytest.mark.xfail  # TODO our test facilities don't support case change?
    @pytest.mark.complete("ssh -vo userknownhostsf")
    def test_5(self, completion):
        assert "UserKnownHostsFile=" in completion

    @pytest.mark.complete("ssh -", require_cmd=True)
    def test_6(self, completion):
        assert completion

    @pytest.mark.complete("ssh -F")
    def test_capital_f_without_space(self, completion):
        assert completion
        assert not any(
            "option requires an argument -- F" in x for x in completion
        )

    @pytest.mark.complete("ssh -F nonexistent ")
    def test_capital_f_nonexistent(self, completion):
        assert completion

    def test_partial_hostname(self, bash, known_hosts):
        first_char, partial_hosts = partialize(bash, known_hosts)
        completion = assert_complete(bash, "ssh %s" % first_char)
        if len(completion) == 1:
            assert completion == partial_hosts[0][1:]
        else:
            assert completion == sorted(x for x in partial_hosts)

    @pytest.mark.parametrize("protocol", "4 6 9".split())
    def test_protocol_option_bundling(self, bash, protocol):
        completion = assert_complete(bash, "ssh -%sF ssh/" % protocol)
        assert "config" in completion

    @pytest.mark.complete("ssh -F config -o ForwardX11=yes ls", cwd="ssh")
    def test_options_with_args_and_arg_counting(self, completion):
        """Options with arguments should not confuse arg counting."""
        assert completion == "_known_host"



import difflib
import os
import re
import shlex
import shutil
import subprocess
import sys
import tempfile
import time
from enum import Enum
from pathlib import Path
from types import TracebackType
from typing import (
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    TextIO,
    Tuple,
    Type,
)

import pexpect  # type: ignore[import]
import pytest

PS1 = "/@"
MAGIC_MARK = "__MaGiC-maRKz-NEtXZVZfKC__"
MAGIC_MARK2 = "Re8SCgEdfN"


def find_unique_completion_pair(
    items: Iterable[str],
) -> Optional[Tuple[str, str]]:
    result = None
    bestscore = 0
    sitems = sorted(set(items))
    for i in range(len(sitems)):
        cur = sitems[i]
        curlen = len(cur)
        prv = sitems[i - 1] if i != 0 else ""
        prvlen = len(prv)
        nxt = sitems[i + 1] if i < len(sitems) - 1 else ""
        nxtlen = len(nxt)
        diffprv = prv == ""
        diffnxt = nxt == ""
        # Analyse each item of the list and look for the minimum length of the
        # partial prefix which is distinct from both nxt and prv. The list
        # is sorted so the prefix will be unique in the entire list.
        for j in range(curlen):
            curchar = cur[j]
            if not diffprv and (j >= prvlen or prv[j] != curchar):
                diffprv = True
            if not diffnxt and (j >= nxtlen or nxt[j] != curchar):
                diffnxt = True
            if diffprv and diffnxt:
                break
        # At the end of the loop, j is the index of last character of
        # the unique partial prefix. The length is one plus that.
        parlen = j + 1
        if parlen >= curlen:
            continue
        # Try to find the most "readable pair"; look for a long pair where
        # part is about half of full.
        if parlen < curlen / 2:
            parlen = int(curlen / 2)
        score = curlen - parlen
        if score > bestscore:
            bestscore = score
            result = (cur[:parlen], cur)
    return result


@pytest.fixture(scope="class")
def output_sort_uniq(bash: pexpect.spawn) -> Callable[[str], List[str]]:
    def _output_sort_uniq(command: str) -> List[str]:
        return sorted(
            set(  # weed out possible duplicates
                assert_bash_exec(bash, command, want_output=True).split()
            )
        )

    return _output_sort_uniq


@pytest.fixture(scope="class")
def part_full_user(
    bash: pexpect.spawn, output_sort_uniq: Callable[[str], List[str]]
) -> Optional[Tuple[str, str]]:
    res = output_sort_uniq("compgen -u")
    pair = find_unique_completion_pair(res)
    if not pair:
        pytest.skip("No suitable test user found")
    return pair


@pytest.fixture(scope="class")
def part_full_group(
    bash: pexpect.spawn, output_sort_uniq: Callable[[str], List[str]]
) -> Optional[Tuple[str, str]]:
    res = output_sort_uniq("compgen -g")
    pair = find_unique_completion_pair(res)
    if not pair:
        pytest.skip("No suitable test user found")
    return pair


@pytest.fixture(scope="class")
def hosts(bash: pexpect.spawn) -> List[str]:
    output = assert_bash_exec(bash, "compgen -A hostname", want_output=True)
    return sorted(set(output.split() + _avahi_hosts(bash)))


@pytest.fixture(scope="class")
def avahi_hosts(bash: pexpect.spawn) -> List[str]:
    return _avahi_hosts(bash)


def _avahi_hosts(bash: pexpect.spawn) -> List[str]:
    output = assert_bash_exec(
        bash,
        "! type avahi-browse &>/dev/null || "
        "avahi-browse -cpr _workstation._tcp 2>/dev/null "
        "| command grep ^= | cut -d';' -f7",
        want_output=None,
    )
    return sorted(set(output.split()))


@pytest.fixture(scope="class")
def known_hosts(bash: pexpect.spawn) -> List[str]:
    output = assert_bash_exec(
        bash,
        '_comp_compgen_known_hosts ""; '
        r'printf "%s\n" "${COMPREPLY[@]}"; unset -v COMPREPLY',
        want_output=True,
    )
    return sorted(set(output.split()))


@pytest.fixture(scope="class")
def user_home(bash: pexpect.spawn) -> Tuple[str, str]:
    user = assert_bash_exec(
        bash, 'id -un 2>/dev/null || echo "$USER"', want_output=True
    ).strip()
    # We used to echo $HOME here, but we expect that it will be consistent with
    # ~user as far as bash is concerned which may not hold.
    home = assert_bash_exec(bash, "echo ~%s" % user, want_output=True).strip()
    return (user, home)


def partialize(
    bash: pexpect.spawn, items: Iterable[str]
) -> Tuple[str, List[str]]:
    """
    Get list of items starting with the first char of first of items.

    Disregard items starting with a COMP_WORDBREAKS character
    (e.g. a colon ~ IPv6 address), they are special cases requiring
    special tests.
    """
    first_char = None
    comp_wordbreaks = assert_bash_exec(
        bash,
        'printf "%s" "$COMP_WORDBREAKS"',
        want_output=True,
        want_newline=False,
    )
    partial_items = []
    for item in sorted(items):
        if first_char is None:
            if item[0] not in comp_wordbreaks:
                first_char = item[0]
                partial_items.append(item)
        elif item.startswith(first_char):
            partial_items.append(item)
        else:
            break
    if first_char is None:
        pytest.skip("Could not generate partial items list from %s" % items)
        # superfluous/dead code to assist mypy; pytest.skip always raises
        assert first_char is not None
    return first_char, partial_items


@pytest.fixture(scope="class")
def bash(request) -> pexpect.spawn:
    logfile: Optional[TextIO] = None
    histfile = None
    tmpdir = None
    bash = None

    if os.environ.get("BASH_COMPLETION_TEST_LOGFILE"):
        logfile = open(os.environ["BASH_COMPLETION_TEST_LOGFILE"], "w")
    elif os.environ.get("CI"):
        logfile = sys.stdout

    testdir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), os.pardir)
    )

    # Create an empty temporary file for HISTFILE.
    #
    # To prevent the tested Bash processes from writing to the user's
    # history file or any other files, we prepare an empty temporary
    # file for each test.
    #
    # - Note that HISTFILE=/dev/null may not work.  It results in the
    #   removal of the device /dev/null and the creation of a regular
    #   file at /dev/null when the number of commands reach
    #   HISTFILESIZE due to a bug in bash 4.3.  This causes execution of
    #   garbage through BASH_COMPLETION_USER_FILE=/dev/null.
    # - Note also that "unset -v HISTFILE" in "test/config/bashrc" was not
    #   adopted because "test/config/bashrc" is loaded after the
    #   history is read from the history file.
    #
    histfile = tempfile.NamedTemporaryFile(
        prefix="bash-completion-test_", delete=False
    )

    try:
        # release the file handle so that Bash can open the file.
        histfile.close()

        env = os.environ.copy()
        env.update(
            dict(
                SRCDIR=testdir,  # TODO needed at least by bashrc
                SRCDIRABS=testdir,
                PS1=PS1,
                INPUTRC="%s/config/inputrc" % testdir,
                TERM="dumb",
                LC_COLLATE="C",  # to match Python's default locale unaware sort
                HISTFILE=histfile.name,
            )
        )

        marker = request.node.get_closest_marker("bashcomp")

        # Set up the current working directory
        cwd = None
        if marker:
            if "cwd" in marker.kwargs and marker.kwargs.get("cwd") is not None:
                cwd = os.path.join(
                    testdir, "fixtures", marker.kwargs.get("cwd")
                )
            elif "temp_cwd" in marker.kwargs and marker.kwargs.get("temp_cwd"):
                tmpdir = tempfile.TemporaryDirectory(
                    prefix="bash-completion-test_"
                )
                cwd = tmpdir.name
        if cwd is None:
            cwd = os.path.join(testdir, "fixtures")
        os.chdir(cwd)

        # Start bash
        bash = pexpect.spawn(
            "%s --norc" % os.environ.get("BASH_COMPLETION_TEST_BASH", "bash"),
            maxread=os.environ.get(
                "BASH_COMPLETION_TEST_PEXPECT_MAXREAD", 20000
            ),
            logfile=logfile,
            cwd=cwd,
            env=env,
            encoding="utf-8",  # TODO? or native or...?
            # FIXME: Tests shouldn't depend on dimensions, but it's difficult to
            # expect robustly enough for Bash to wrap lines anywhere (e.g. inside
            # MAGIC_MARK).  Increase window width to reduce wrapping.
            dimensions=(24, 240),
            # TODO? codec_errors="replace",
        )
        bash.expect_exact(PS1)

        # Load bashrc and bash_completion
        bash_completion = os.environ.get(
            "BASH_COMPLETION_TEST_BASH_COMPLETION",
            "%s/../bash_completion" % testdir,
        )
        assert_bash_exec(bash, "source '%s/config/bashrc'" % testdir)
        assert_bash_exec(bash, "source '%s'" % bash_completion)

        # Use command name from marker if set, or grab from test filename
        cmd = None  # type: Optional[str]
        cmd_found = False
        if marker:
            cmd = marker.kwargs.get("cmd")
            cmd_found = "cmd" in marker.kwargs
            # Run pre-test commands, early so they're usable in skipif
            for pre_cmd in marker.kwargs.get("pre_cmds", []):
                assert_bash_exec(bash, pre_cmd, want_output=None)
            # Process skip and xfail conditions
            skipif = marker.kwargs.get("skipif")
            if skipif:
                try:
                    assert_bash_exec(bash, skipif, want_output=None)
                except AssertionError:
                    pass
                else:
                    bash.close()
                    bash = None
                    pytest.skip(skipif)
            xfail = marker.kwargs.get("xfail")
            if xfail:
                try:
                    assert_bash_exec(bash, xfail, want_output=None)
                except AssertionError:
                    pass
                else:
                    pytest.xfail(xfail)
        if not cmd_found:
            match = re.search(
                r"^test_(.+)\.py$", os.path.basename(str(request.fspath))
            )
            if match:
                cmd = match.group(1)
        if (
            marker
            and marker.kwargs
            and marker.kwargs.get("require_cmd", False)
        ):
            if not is_bash_type(bash, cmd):
                pytest.skip("Command not found")

        request.cls.cmd = cmd

        if (cmd_found and cmd is None) or is_testable(bash, cmd):
            before_env = get_env(bash)
            yield bash
            # Not exactly sure why, but some errors leave bash in state where
            # getting the env here would fail and trash our test output. So
            # reset to a good state first (Ctrl+C, expect prompt).
            bash.sendintr()
            bash.expect_exact(PS1)
            diff_env(
                before_env,
                get_env(bash),
                marker.kwargs.get("ignore_env") if marker else "",
            )

        if marker:
            for post_cmd in marker.kwargs.get("post_cmds", []):
                assert_bash_exec(bash, post_cmd, want_output=None)

    finally:
        # Clean up
        if bash:
            bash.close()
        if tmpdir:
            tmpdir.cleanup()
        if histfile:
            try:
                os.remove(histfile.name)
            except OSError:
                pass
        if logfile and logfile != sys.stdout:
            logfile.close()


def is_testable(bash: pexpect.spawn, cmd: Optional[str]) -> bool:
    if not cmd:
        pytest.fail("Could not resolve name of command to test")
        return False
    if not load_completion_for(bash, cmd):
        pytest.skip("No completion for command %s" % cmd)
    return True


def is_bash_type(bash: pexpect.spawn, cmd: Optional[str]) -> bool:
    if not cmd:
        return False
    typecmd = "type %s &>/dev/null && echo -n 0 || echo -n 1" % cmd
    bash.sendline(typecmd)
    bash.expect_exact(typecmd + "\r\n")
    result = bash.expect_exact(["0", "1"]) == 0
    bash.expect_exact(PS1)
    return result


def load_completion_for(bash: pexpect.spawn, cmd: str) -> bool:
    try:
        # Allow _comp_load to fail so we can test completions
        # that are directly loaded in bash_completion without a separate file.
        assert_bash_exec(bash, "_comp_load -- %s || :" % cmd)
        assert_bash_exec(bash, "complete -p -- %s &>/dev/null" % cmd)
    except AssertionError:
        return False
    return True


def assert_bash_exec(
    bash: pexpect.spawn,
    cmd: str,
    want_output: Optional[bool] = False,
    want_newline=True,
) -> str:
    """
    :param want_output: if None, don't care if got output or not
    """

    # Send command
    bash.sendline(cmd)
    bash.expect_exact(cmd)

    # Find prompt, output is before it
    bash.expect_exact("%s%s" % ("\r\n" if want_newline else "", PS1))
    output = bash.before

    # Retrieve exit status
    echo = "echo $?"
    bash.sendline(echo)
    got = bash.expect(
        [
            r"^%s\r\n(\d+)\r\n%s" % (re.escape(echo), re.escape(PS1)),
            PS1,
            pexpect.EOF,
            pexpect.TIMEOUT,
        ]
    )
    status = bash.match.group(1) if got == 0 else "unknown"

    assert status == "0", 'Error running "%s": exit status=%s, output="%s"' % (
        cmd,
        status,
        output,
    )
    if want_output is not None:
        if output:
            assert want_output, (
                'Unexpected output from "%s": exit status=%s, output="%s"'
                % (
                    cmd,
                    status,
                    output,
                )
            )
        else:
            assert not want_output, (
                'Expected output from "%s": exit status=%s, output="%s"'
                % (
                    cmd,
                    status,
                    output,
                )
            )

    return output


class bash_env_saved:
    counter: int = 0

    class saved_state(Enum):
        ChangesDetected = 1
        ChangesIgnored = 2

    def __init__(self, bash: pexpect.spawn, sendintr: bool = False):
        bash_env_saved.counter += 1
        self.prefix: str = "_comp__test_%d" % bash_env_saved.counter

        self.bash = bash
        self.cwd_changed: bool = False
        self.saved_set: Dict[str, bash_env_saved.saved_state] = {}
        self.saved_shopt: Dict[str, bash_env_saved.saved_state] = {}
        self.saved_variables: Dict[str, bash_env_saved.saved_state] = {}
        self.sendintr = sendintr

        self.noexcept: bool = False
        self.captured_error: Optional[Exception] = None

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> None:
        self._restore_env()
        return None

    def _safe_sendintr(self):
        try:
            self.bash.sendintr()
            self.bash.expect_exact(PS1)
        except Exception as e:
            if self.noexcept:
                self.captured_error = e
            else:
                raise

    def _safe_exec(self, cmd: str):
        try:
            self.bash.sendline(cmd)
            self.bash.expect_exact(cmd)
            self.bash.expect_exact("\r\n" + PS1)
        except Exception as e:
            if self.noexcept:
                self._safe_sendintr()
                self.captured_error = e
            else:
                raise

    def _safe_assert(self, cmd: str):
        try:
            assert_bash_exec(self.bash, cmd, want_output=None)
        except Exception as e:
            if self.noexcept:
                self._safe_sendintr()
                self.captured_error = e
            else:
                raise

    def _copy_variable(self, src_var: str, dst_var: str):
        self._safe_exec(
            "if [[ ${%s+set} ]]; then %s=${%s}; else unset -v %s; fi"
            % (src_var, dst_var, src_var, dst_var),
        )

    def _unset_variable(self, varname: str):
        self._safe_exec("unset -v %s" % varname)

    def _save_cwd(self):
        if not self.cwd_changed:
            self.cwd_changed = True
            self._copy_variable("PWD", "%s_OLDPWD" % self.prefix)

    def _check_set(self, name: str):
        if self.saved_set[name] != bash_env_saved.saved_state.ChangesDetected:
            return
        self._safe_assert(
            '[[ $(shopt -po %s) == "${%s_NEWSHOPT_%s}" ]]'
            % (name, self.prefix, name),
        )

    def _unprotect_set(self, name: str):
        if name not in self.saved_set:
            self.saved_set[name] = bash_env_saved.saved_state.ChangesDetected
            self._safe_exec(
                "%s_OLDSHOPT_%s=$(shopt -po %s || true)"
                % (self.prefix, name, name),
            )
        else:
            self._check_set(name)

    def _protect_set(self, name: str):
        self._safe_exec(
            "%s_NEWSHOPT_%s=$(shopt -po %s || true)"
            % (self.prefix, name, name),
        )

    def _check_shopt(self, name: str):
        if (
            self.saved_shopt[name]
            != bash_env_saved.saved_state.ChangesDetected
        ):
            return
        self._safe_assert(
            '[[ $(shopt -p %s) == "${%s_NEWSHOPT_%s}" ]]'
            % (name, self.prefix, name),
        )

    def _unprotect_shopt(self, name: str):
        if name not in self.saved_shopt:
            self.saved_shopt[name] = bash_env_saved.saved_state.ChangesDetected
            self._safe_exec(
                "%s_OLDSHOPT_%s=$(shopt -p %s || true)"
                % (self.prefix, name, name),
            )
        else:
            self._check_shopt(name)

    def _protect_shopt(self, name: str):
        self._safe_exec(
            "%s_NEWSHOPT_%s=$(shopt -p %s || true)"
            % (self.prefix, name, name),
        )

    def _check_variable(self, varname: str):
        if (
            self.saved_variables[varname]
            != bash_env_saved.saved_state.ChangesDetected
        ):
            return
        try:
            self._safe_assert(
                '[[ ${%s-%s} == "${%s_NEWVAR_%s-%s}" ]]'
                % (varname, MAGIC_MARK2, self.prefix, varname, MAGIC_MARK2),
            )
        except Exception:
            self._copy_variable(
                "%s_NEWVAR_%s" % (self.prefix, varname), varname
            )
            raise
        else:
            if self.noexcept and self.captured_error:
                self._copy_variable(
                    "%s_NEWVAR_%s" % (self.prefix, varname), varname
                )

    def _unprotect_variable(self, varname: str):
        if varname not in self.saved_variables:
            self.saved_variables[varname] = (
                bash_env_saved.saved_state.ChangesDetected
            )
            self._copy_variable(
                varname, "%s_OLDVAR_%s" % (self.prefix, varname)
            )
        else:
            self._check_variable(varname)

    def _protect_variable(self, varname: str):
        self._copy_variable(varname, "%s_NEWVAR_%s" % (self.prefix, varname))

    def _restore_env(self):
        self.noexcept = True

        if self.sendintr:
            self._safe_sendintr()

        # We first go back to the original directory before restoring
        # variables because "cd" affects "OLDPWD".
        if self.cwd_changed:
            self._unprotect_variable("OLDPWD")
            self._safe_exec('command cd -- "$%s_OLDPWD"' % self.prefix)
            self._protect_variable("OLDPWD")
            self._unset_variable("%s_OLDPWD" % self.prefix)
            self.cwd_changed = False

        for varname in self.saved_variables:
            self._check_variable(varname)
            self._copy_variable(
                "%s_OLDVAR_%s" % (self.prefix, varname), varname
            )
            self._unset_variable("%s_OLDVAR_%s" % (self.prefix, varname))
            self._unset_variable("%s_NEWVAR_%s" % (self.prefix, varname))
        self.saved_variables = {}

        for name in self.saved_shopt:
            self._check_shopt(name)
            self._safe_exec('eval "$%s_OLDSHOPT_%s"' % (self.prefix, name))
            self._unset_variable("%s_OLDSHOPT_%s" % (self.prefix, name))
            self._unset_variable("%s_NEWSHOPT_%s" % (self.prefix, name))
        self.saved_shopt = {}

        for name in self.saved_set:
            self._check_set(name)
            self._safe_exec('eval "$%s_OLDSHOPT_%s"' % (self.prefix, name))
            self._unset_variable("%s_OLDSHOPT_%s" % (self.prefix, name))
            self._unset_variable("%s_NEWSHOPT_%s" % (self.prefix, name))
        self.saved_set = {}

        self.noexcept = False
        if self.captured_error:
            raise self.captured_error

    def chdir(self, path: str):
        self._save_cwd()
        self.cwd_changed = True
        self._unprotect_variable("OLDPWD")
        self._safe_exec("command cd -- %s" % shlex.quote(path))
        self._protect_variable("OLDPWD")

    def set(self, name: str, value: bool):
        self._unprotect_set(name)
        if value:
            self._safe_exec("set -o %s" % name)
        else:
            self._safe_exec("set +o %s" % name)
        self._protect_set(name)

    def save_set(self, name: str):
        self._unprotect_set(name)
        self.saved_set[name] = bash_env_saved.saved_state.ChangesIgnored

    def shopt(self, name: str, value: bool):
        self._unprotect_shopt(name)
        if value:
            self._safe_exec("shopt -s %s" % name)
        else:
            self._safe_exec("shopt -u %s" % name)
        self._protect_shopt(name)

    def save_shopt(self, name: str):
        self._unprotect_shopt(name)
        self.saved_shopt[name] = bash_env_saved.saved_state.ChangesIgnored

    def write_variable(self, varname: str, new_value: str, quote: bool = True):
        if quote:
            new_value = shlex.quote(new_value)
        self._unprotect_variable(varname)
        self._safe_exec("%s=%s" % (varname, new_value))
        self._protect_variable(varname)

    def save_variable(self, varname: str):
        self._unprotect_variable(varname)
        self.saved_variables[varname] = (
            bash_env_saved.saved_state.ChangesIgnored
        )

    # TODO: We may restore the "export" attribute as well though it is
    #   not currently tested in "diff_env"
    def write_env(self, envname: str, new_value: str, quote: bool = True):
        if quote:
            new_value = shlex.quote(new_value)
        self._unprotect_variable(envname)
        self._safe_exec("export %s=%s" % (envname, new_value))
        self._protect_variable(envname)


def get_env(bash: pexpect.spawn) -> List[str]:
    return [
        x
        for x in assert_bash_exec(
            bash,
            "_comp__test_get_env",
            want_output=True,
        )
        .strip()
        .splitlines()
        # Sometimes there are empty lines in the output due to unknown
        # reasons, e.g. in GitHub Actions' macos-latest OS. Filter them out.
        if x
    ]


def diff_env(before: List[str], after: List[str], ignore: str):
    diff = [
        x
        for x in difflib.unified_diff(before, after, n=0, lineterm="")
        # Remove unified diff markers:
        if not re.search(r"^(---|\+\+\+|@@ )", x)
        # Ignore variables expected to change:
        and not re.search(
            r"^[-+](_|PPID|BASH_REMATCH|(BASH_)?LINENO)=",
            x,
            re.ASCII,
        )
        # Ignore likely completion functions added by us:
        and not re.search(r"^\+declare -f _.+", x)
        # ...and additional specified things:
        and not re.search(ignore or "^$", x)
    ]
    # For some reason, COMP_WORDBREAKS gets added to the list after
    # saving. Remove its changes, and note that it may take two lines.
    for i in range(0, len(diff)):
        if re.match("^[-+]COMP_WORDBREAKS=", diff[i]):
            if i < len(diff) and not re.match(r"^\+[\w]+=", diff[i + 1]):
                del diff[i + 1]
            del diff[i]
            break
    assert not diff, "Environment should not be modified"


class CompletionResult(Iterable[str]):
    """
    Class to hold completion results.
    """

    def __init__(self, output: Optional[str] = None):
        """
        :param output: All completion output as-is.
        """
        self.output = output or ""

    def endswith(self, suffix: str) -> bool:
        return self.output.endswith(suffix)

    def startswith(self, prefix: str) -> bool:
        return self.output.startswith(prefix)

    def _items(self) -> List[str]:
        return [x.strip() for x in self.output.strip().splitlines()]

    def __eq__(self, expected: object) -> bool:
        """
        Returns True if completion contains expected items, and no others.

        Defining __eq__ this way is quite ugly, but facilitates concise
        testing code.
        """
        if isinstance(expected, str):
            expiter = [expected]  # type: Iterable
        elif not isinstance(expected, Iterable):
            return False
        else:
            expiter = expected
        return self._items() == expiter

    def __contains__(self, item: str) -> bool:
        return item in self._items()

    def __iter__(self) -> Iterator[str]:
        return iter(self._items())

    def __len__(self) -> int:
        return len(self._items())

    def __repr__(self) -> str:
        return "<CompletionResult %s>" % self._items()


def assert_complete(
    bash: pexpect.spawn, cmd: str, **kwargs
) -> CompletionResult:
    skipif = kwargs.get("skipif")
    if skipif:
        try:
            assert_bash_exec(bash, skipif, want_output=None)
        except AssertionError:
            pass
        else:
            pytest.skip(skipif)
    xfail = kwargs.get("xfail")
    if xfail:
        try:
            assert_bash_exec(bash, xfail, want_output=None)
        except AssertionError:
            pass
        else:
            pytest.xfail(xfail)

    with bash_env_saved(bash, sendintr=True) as bash_env:
        cwd = kwargs.get("cwd")
        if cwd:
            bash_env.chdir(str(cwd))

        for k, v in kwargs.get("env", {}).items():
            bash_env.write_env(k, v, quote=False)

        for k, v in kwargs.get("shopt", {}).items():
            bash_env.shopt(k, v)

        input_cmd = cmd
        rendered_cmd = kwargs.get("rendered_cmd", cmd)
        re_MAGIC_MARK = re.escape(MAGIC_MARK)

        trail = kwargs.get("trail")
        if trail:
            # \002 = ^B = cursor left
            input_cmd += trail + "\002" * len(trail)
            rendered_cmd += trail + "\b" * len(trail)

            # After reading the results, something weird happens. For most test
            # setups, as expected (pun intended!), MAGIC_MARK follows as
            # is. But for some others (e.g. CentOS 6, Ubuntu 14 test
            # containers), we get MAGIC_MARK one character a time, followed
            # each time by trail and the corresponding number of \b's. Don't
            # know why, but accept it until/if someone finds out.  Or just be
            # fine with it indefinitely, the visible and practical end result
            # on a terminal is the same anyway.
            maybe_trail = "(%s%s)?" % (re.escape(trail), "\b" * len(trail))
            re_MAGIC_MARK = "".join(
                re.escape(x) + maybe_trail for x in MAGIC_MARK
            )

        bash.send(input_cmd + "\t")
        # Sleep a bit if requested, to avoid `.*` matching too early
        time.sleep(kwargs.get("sleep_after_tab", 0))
        bash.expect_exact(rendered_cmd)
        bash.send(MAGIC_MARK)
        got = bash.expect(
            [
                # 0: multiple lines, result in .before
                r"\r\n" + re.escape(PS1 + rendered_cmd) + ".*" + re_MAGIC_MARK,
                # 1: no completion
                r"^" + re_MAGIC_MARK,
                # 2: on same line, result in .match
                r"^([^\r]+)%s$" % re_MAGIC_MARK,
                # 3: error messages
                r"^([^\r].*)%s$" % re_MAGIC_MARK,
                pexpect.EOF,
                pexpect.TIMEOUT,
            ]
        )
        if got == 0:
            output = re.sub(re_MAGIC_MARK + "$", "", bash.before)
            return CompletionResult(output)
        elif got == 2:
            output = bash.match.group(1)
            return CompletionResult(output)
        elif got == 3:
            output = bash.match.group(1)
            raise AssertionError("Unexpected output: [%s]" % output)
        else:
            # TODO: warn about EOF/TIMEOUT?
            return CompletionResult()


@pytest.fixture
def completion(request, bash: pexpect.spawn) -> CompletionResult:
    marker = request.node.get_closest_marker("complete")
    if not marker:
        return CompletionResult()
    for pre_cmd in marker.kwargs.get("pre_cmds", []):
        assert_bash_exec(bash, pre_cmd, want_output=None)
    cmd = getattr(request.cls, "cmd", None)
    if marker.kwargs.get("require_longopt"):
        # longopt completions require both command presence and that it
        # responds something useful to --help
        if "require_cmd" not in marker.kwargs:
            marker.kwargs["require_cmd"] = True
        if "xfail" not in marker.kwargs:
            marker.kwargs["xfail"] = (
                # --help is required to exit with zero in order to not get a
                # positive for cases where it errors out with a message like
                # "foo: unrecognized option '--help'"
                "! %s --help &>/dev/null || "
                "! %s --help 2>&1 | command grep -qF -- --help"
            ) % ((cmd,) * 2)
    if marker.kwargs.get("require_cmd") and not is_bash_type(bash, cmd):
        pytest.skip("Command not found")

    return assert_complete(bash, marker.args[0], **marker.kwargs)


def in_container() -> bool:
    try:
        container = subprocess.check_output(
            "virt-what || systemd-detect-virt --container",
            stderr=subprocess.DEVNULL,
            shell=True,
        ).strip()
    except subprocess.CalledProcessError:
        container = b""
    if container and container != b"none":
        return True
    if os.path.exists("/.dockerenv"):
        return True
    try:
        with open("/proc/1/environ", "rb") as f:
            # LXC, others?
            if any(
                x.startswith(b"container=") for x in f.readline().split(b"\0")
            ):
                return True
    except OSError:
        pass
    return False


def prepare_fixture_dir(
    request, files: Iterable[str], dirs: Iterable[str]
) -> Tuple[Path, List[str], List[str]]:
    """
    Fixture to prepare a test dir with dummy contents on the fly.

    Tests that contain filenames differing only by case should use this to
    prepare a dir on the fly rather than including their fixtures in git and
    the tarball. This is to work better with case insensitive file systems.
    """
    tempdir = Path(tempfile.mkdtemp(prefix="bash-completion-fixture-dir"))
    request.addfinalizer(lambda: shutil.rmtree(str(tempdir)))

    old_cwd = os.getcwd()
    try:
        os.chdir(tempdir)
        new_files, new_dirs = create_dummy_filedirs(files, dirs)
    finally:
        os.chdir(old_cwd)

    return tempdir, new_files, new_dirs


def create_dummy_filedirs(
    files: Iterable[str], dirs: Iterable[str]
) -> Tuple[List[str], List[str]]:
    """
    Create dummy files and directories on the fly in the current directory.

    Tests that contain filenames differing only by case should use this to
    prepare a dir on the fly rather than including their fixtures in git and
    the tarball. This is to work better with case insensitive file systems.
    """
    new_files = []
    new_dirs = []

    for dir_ in dirs:
        path = Path(dir_)
        if not path.exists():
            path.mkdir()
            new_dirs.append(dir_)
    for file_ in files:
        path = Path(file_)
        if not path.exists():
            path.touch()
            new_files.append(file_)

    return sorted(new_files), sorted(new_dirs)


class TestUnitBase:
    def _test_unit(
        self, func, bash, comp_words, comp_cword, comp_line, comp_point, arg=""
    ):
        assert_bash_exec(
            bash,
            "COMP_WORDS=%s COMP_CWORD=%d COMP_LINE=%s COMP_POINT=%d"
            % (comp_words, comp_cword, shlex.quote(comp_line), comp_point),
        )
        output = assert_bash_exec(bash, func % arg, want_output=True)
        return output.strip()



import pytest

from conftest import assert_bash_exec


@pytest.mark.bashcomp(cmd=None, ignore_env=r"^\+COMPREPLY=")
class TestUnitCompgenXinetdServices:
    def test_direct(self, bash):
        assert_bash_exec(bash, "_comp_compgen_xinetd_services >/dev/null")

    def test_env_non_pollution(self, bash):
        """Test environment non-pollution, detected at teardown."""
        assert_bash_exec(
            bash, "foo() { _comp_compgen_xinetd_services; }; foo; unset -f foo"
        )

    def test_basic(self, bash):
        output = assert_bash_exec(
            bash,
            "foo() { local _comp__test_xinetd_dir=$PWD/_comp_compgen_xinetd_services/xinetd.d; unset -v COMPREPLY; "
            '_comp_compgen_xinetd_services; printf "%s\\n" "${COMPREPLY[@]}"; }; foo; unset -f foo',
            want_output=True,
        )
        assert sorted(output.split()) == ["arp", "ifconfig"]



import pytest


class TestIpcalc:
    @pytest.mark.complete("ipcalc -", require_cmd=True)
    def test_options(self, completion):
        assert any(x in completion for x in "--help -h".split())

    @pytest.mark.complete("ipcalc --split -")
    def test_split_3args_1(self, completion):
        assert not completion

    @pytest.mark.complete("ipcalc --split 1 -")
    def test_split_3args_2(self, completion):
        assert not completion

    @pytest.mark.complete("ipcalc --split 1 2 -")
    def test_split_3args_3(self, completion):
        assert not completion

    @pytest.mark.complete("ipcalc --split 1 2 3 -", require_cmd=True)
    def test_split_3args_4(self, completion):
        assert any(x in completion for x in "--help -h".split())



import pytest

from conftest import assert_bash_exec, bash_env_saved


@pytest.mark.bashcomp(cmd=None, ignore_env=r"^\+COMPREPLY=")
class TestUnitCompgenPnames:
    def test_smoke(self, bash):
        with bash_env_saved(bash) as bash_env:
            bash_env.write_variable("cur", "")
            assert_bash_exec(bash, "_comp_compgen_pnames >/dev/null")

    def test_non_pollution(self, bash):
        """Test environment non-pollution, detected at teardown."""
        assert_bash_exec(
            bash,
            "foo() { local cur=; _comp_compgen_pnames; }; foo; unset -f foo",
        )

    def test_something(self, bash):
        """Test that we get something."""
        with bash_env_saved(bash) as bash_env:
            bash_env.write_variable("cur", "")
            completion = assert_bash_exec(
                bash,
                r'_comp_compgen_pids; printf "%s\n" "${COMPREPLY[@]}"',
                want_output=True,
            ).split()
        assert completion



import pytest

from conftest import assert_bash_exec


@pytest.mark.bashcomp(cmd=None)
class TestUtilCompgenAvailableInterfaces:
    @pytest.fixture
    def functions(self, bash):
        assert_bash_exec(
            bash,
            "_comp__test_dump() { ((${#arr[@]})) && printf '<%s>' \"${arr[@]}\"; echo; }",
        )
        assert_bash_exec(
            bash,
            '_comp__test_compgen() { local -a arr=(00); _comp_compgen -v arr "$@"; _comp__test_dump; }',
        )

    def test_1_trailing_colons(self, bash, functions):
        output = assert_bash_exec(
            bash,
            "_comp__test_compgen available_interfaces",
            want_output=True,
        )
        assert ":>" not in output.strip()



import pytest


class TestVgdisplay:
    @pytest.mark.complete(
        "vgdisplay -", require_cmd=True, xfail="! vgdisplay --help &>/dev/null"
    )
    def test_1(self, completion):
        assert completion



import pytest

from conftest import TestUnitBase, assert_bash_exec


@pytest.mark.bashcomp(
    cmd=None,
    ignore_env=r"^[+-](REPLY|cword|words|COMP_(WORDS|CWORD|LINE|POINT))=",
)
class TestUnitCountArgs(TestUnitBase):
    @pytest.fixture
    def functions(self, bash):
        assert_bash_exec(
            bash,
            '_comp__test_unit() { local -a words=(); local cword REPLY=""; _comp__reassemble_words "<>&" words cword; _comp_count_args "$@"; echo "$REPLY"; }',
        )

    def _test(self, *args, **kwargs):
        return self._test_unit("_comp__test_unit %s", *args, **kwargs)

    def test_1(self, bash):
        assert_bash_exec(
            bash,
            'COMP_LINE= COMP_POINT=0 COMP_WORDS=() COMP_CWORD=; _comp_count_args -n ""',
        )

    def test_2(self, bash, functions):
        """a b| should set args to 1"""
        output = self._test(bash, "(a b)", 1, "a b", 3)
        assert output == "1"

    def test_3(self, bash, functions):
        """a b|c should set args to 1"""
        output = self._test(bash, "(a bc)", 1, "a bc", 3)
        assert output == "1"

    def test_4(self, bash, functions):
        """a b c| should set args to 2"""
        output = self._test(bash, "(a b c)", 2, "a b c", 4)
        assert output == "2"

    def test_5(self, bash, functions):
        """a b| c should set args to 1"""
        output = self._test(bash, "(a b c)", 1, "a b c", 3)
        assert output == "1"

    def test_6(self, bash, functions):
        """a b -c| d should set args to 2"""
        output = self._test(bash, "(a b -c d)", 2, "a b -c d", 6)
        assert output == "2"

    def test_7(self, bash, functions):
        """a b -c d e| with -c arg excluded should set args to 2"""
        output = self._test(
            bash, "(a b -c d e)", 4, "a b -c d e", 10, arg='-a "@(-c|--foo)"'
        )
        assert output == "2"

    def test_8(self, bash, functions):
        """a -b -c d e| with -c arg excluded
        and -b included should set args to 1"""
        output = self._test(
            bash,
            "(a -b -c d e)",
            4,
            "a -b -c d e",
            11,
            arg='-a "@(-c|--foo)" -i "-[b]"',
        )
        assert output == "2"

    def test_9(self, bash, functions):
        """a -b -c d e| with -b included should set args to 3"""
        output = self._test(
            bash, "(a -b -c d e)", 4, "a -b -c d e", 11, arg='-i "-b"'
        )
        assert output == "3"

    def test_10_single_hyphen_1(self, bash):
        """- should be counted as an argument representing stdout/stdin"""
        output = self._test(bash, "(a -b - c -d e)", 5, "a -b - c -d e", 12)
        assert output == "3"

    def test_10_single_hyphen_2(self, bash):
        """- in an option argument should be skipped"""
        output = self._test(
            bash, "(a -b - c - e)", 5, "a -b - c - e", 11, arg='-a "-b"'
        )
        assert output == "3"

    def test_11_double_hyphen_1(self, bash):
        """all the words after -- should be counted"""
        output = self._test(
            bash, "(a -b -- -c -d e)", 5, "a -b -- -c -d e", 14
        )
        assert output == "3"

    def test_11_double_hyphen_2(self, bash):
        """all the words after -- should be counted"""
        output = self._test(bash, "(a b -- -c -d e)", 5, "a b -- -c -d e", 13)
        assert output == "4"

    def test_12_exclude_optarg_1(self, bash):
        """an option argument should be skipped even if it matches the argument pattern"""
        output = self._test(
            bash, "(a -o -x b c)", 4, "a -o -x b c", 10, arg='-a "-o" -i "-x"'
        )
        assert output == "2"

    def test_12_exclude_optarg_2(self, bash):
        """an option argument should be skipped even if it matches the argument pattern"""
        output = self._test(
            bash,
            "(a -o -x -x c)",
            4,
            "a -o -x -x c",
            11,
            arg='-a "-o" -i "-x"',
        )
        assert output == "2"

    def test_12_exclude_optarg_3(self, bash):
        """an option argument should be skipped even if it matches the argument pattern"""
        output = self._test(
            bash,
            "(a -o -x -y c)",
            4,
            "a -o -x -y c",
            11,
            arg='-a "-o" -i "-x"',
        )
        assert output == "1"

    def test_13_plus_option_optarg(self, bash):
        """When +o is specified to be an option taking an option argument, it should not be counted as an argument"""
        output = self._test(
            bash, "(a +o b c)", 3, "a +o b c", 7, arg='-a "+o"'
        )
        assert output == "1"

    def test_14_no_optarg_chain_1(self, bash):
        """an option argument should not take another option argument"""
        output = self._test(
            bash, "(a -o -o -o -o c)", 5, "a -o -o -o -o c", 14, arg='-a "-o"'
        )
        assert output == "1"

    def test_14_no_optarg_chain_2(self, bash):
        """an option argument should not take another option argument"""
        output = self._test(
            bash,
            "(a -o -o b -o -o c)",
            6,
            "a -o -o b -o -o c",
            16,
            arg='-a "-o"',
        )
        assert output == "2"

    def test_15_double_hyphen_optarg(self, bash):
        """-- should lose its meaning when it is an option argument"""
        output = self._test(
            bash, "(a -o -- -b -c d)", 5, "a -o -- -b -c d", 14, arg='-a "-o"'
        )
        assert output == "1"

    def test_16_empty_word(self, bash):
        """An empty word should not take an option argument"""
        output = self._test(bash, "(a '' x '' y d)", 5, "a  x  y d", 8)
        assert output == "5"



import pytest

from conftest import find_unique_completion_pair


@pytest.mark.bashcomp(cmd=None)
class TestUnitFindUniqueCompletionPair:
    def _test(self, inp: str, exp: str) -> None:
        res = find_unique_completion_pair(inp.split())
        if exp:
            part, cont = exp.split()
            assert res == (part, part + cont)
        else:
            assert not exp

    def test_1(self):
        self._test("a", "")

    def test_2(self):
        self._test("ab", "a b")

    def test_3(self):
        self._test("ab ab ab", "a b")

    def test_4(self):
        self._test("a ab abcd abc", "")

    def test_5(self):
        self._test("user1 user2", "")

    def test_6(self):
        self._test("root username1 username2", "ro ot")

    def test_7(self):
        self._test("root username21 username2", "ro ot")

    def test_8(self):
        self._test(
            "long_user_name lang_user_name long_usor_name", "lang_us er_name"
        )

    def test_9(self):
        self._test(
            "lang_user_name1 long_user_name lang_user_name long_usor_name",
            "long_use r_name",
        )

    def test_10(self):
        self._test("root username", "user name")

    def test_11(self):
        self._test("a aladin", "ala din")

    def test_12(self):
        self._test("ala aladin", "alad in")



import pytest


class TestLdd:
    @pytest.mark.complete("ldd ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete(
        "ldd -", require_cmd=True, xfail="! ldd --help &>/dev/null"
    )
    def test_options(self, completion):
        assert completion



import pytest

from conftest import assert_bash_exec


class TestExport:
    @pytest.mark.complete("export BASH")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("export -n BASH")
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("export -p ")
    def test_3(self, completion):
        assert not completion

    @pytest.mark.complete("export FOO=", cwd="shared/default")
    def test_4(self, completion):
        assert completion == ["bar", "bar bar.d/", "foo", "foo.d/"]

    @pytest.mark.complete("export FOO=f", cwd="shared/default")
    def test_5(self, completion):
        assert completion == ["foo", "foo.d/"]

    @pytest.mark.complete("export -fn _comp_cmd_ex")
    def test_6(self, completion):
        assert completion == "port"

    @pytest.mark.complete(r"export FOO=$BASH")
    def test_7(self, completion):
        assert completion

    @pytest.mark.complete("export -", require_cmd=True)
    def test_8(self, completion):
        assert completion

    @pytest.fixture(scope="class")
    def export_f_canary(self, request, bash):
        assert_bash_exec(bash, "_comp__test_export_f_canary() { return; }")

    @pytest.mark.complete("export -f _comp__test_export_f_canar")
    def test_no_equals_sign_for_function(self, completion, export_f_canary):
        assert completion
        assert "=" not in "".join(completion)
        assert completion.endswith(" ")



import inspect

import pytest


class TestPytest:
    @pytest.mark.complete("pytest ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("pytest -", require_cmd=True)
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("pytest ../t/test_pytest.py:")
    def test_classes_and_functions(self, completion):
        assert completion == ":TestPytest :test_function_canary".split()

    @pytest.mark.complete("pytest ../t/test_pytest.py::TestPytest::")
    def test_class_methods(self, completion):
        methods = [
            x[0]
            for x in inspect.getmembers(self, predicate=inspect.ismethod)
            if x[0].startswith("test_")
        ]
        assert completion == methods

    @pytest.mark.complete("pytest pytest/test_async.py:")
    def test_classes_and_async_functions(self, completion):
        assert completion == ":Testing :test_positive".split()

    @pytest.mark.complete("pytest pytest/test_async.py::Testing::")
    def test_async_class_methods(self, completion):
        assert completion == "test_positive"

    def non_test_cananary_method(self):
        pass


def test_function_canary():
    pass


def non_test_canary():
    pass


class NonTestCanaryClass:
    def test_is_this_function_not(self):
        pass



import pytest


@pytest.mark.bashcomp(pre_cmds=("PATH=/usr/lib/mailman/bin:$PATH",))
class TestListOwners:
    @pytest.mark.complete("list_owners -")
    def test_1(self, completion):
        assert completion



import pytest


class TestA2x:
    @pytest.mark.complete("a2x ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("a2x -", require_cmd=True)
    def test_2(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(cmd="sbcl-mt")
class TestSbclMt:
    @pytest.mark.complete("sbcl-mt shared/default/")
    def test_1(self, completion):
        assert completion == ["bar", "bar bar.d/", "foo", "foo.d/"]



import fnmatch
import os

import pytest


class TestInstallpkg:
    @pytest.mark.complete("installpkg -")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("installpkg --")
    def test_2(self, completion):
        assert (
            completion
            == "--ask --infobox --md5sum --menu "
            "--priority --root --tagfile --terse --warn".split()
        )

    @pytest.mark.complete("installpkg --root ")
    def test_3(self, completion):
        dirs = sorted(x for x in os.listdir(".") if os.path.isdir("./%s" % x))
        assert completion == ["%s/" % x for x in dirs]

    @pytest.mark.complete("installpkg ", cwd="slackware/home")
    def test_4(self, completion):
        expected = sorted(
            [
                "%s/" % x
                for x in os.listdir("slackware/home")
                if os.path.isdir("./slackware/home/%s" % x)
            ]
            + [
                x
                for x in os.listdir("slackware/home")
                if os.path.isfile("./slackware/home/%s" % x)
                and fnmatch.fnmatch(x, "*.t[bglx]z")
            ]
        )
        assert completion == expected



import pytest

from conftest import assert_complete, create_dummy_filedirs


@pytest.mark.bashcomp(temp_cwd=True)
class TestKpdf:
    def test_1(self, bash):
        files, dirs = create_dummy_filedirs(
            ".eps .EPS .pdf .PDF .ps .PS .txt".split(),
            "foo".split(),
        )

        completion = assert_complete(bash, "kpdf ")
        assert completion == [
            x
            for x in sorted(files + ["%s/" % d for d in dirs])
            if x.lower() != ".txt"
        ]



import pytest


@pytest.mark.bashcomp(
    pre_cmds=("unalias -a", "alias foo=bar", "alias bar='foo foo'"),
    post_cmds=("unalias -a",),
)
class TestAlias:
    @pytest.mark.complete("alias ")
    def test_1(self, completion):
        assert completion == "bar foo".split()

    @pytest.mark.xfail  # TODO: Would like this completion to work
    @pytest.mark.complete("alias foo=")
    def test_2(self, completion):
        assert completion == "foo='bar'"
        assert not completion.endswith(" ")

    @pytest.mark.complete("alias ", trail="foo")
    def test_alias_at_point(self, completion):
        assert completion == "bar foo".split()

    @pytest.mark.complete("alias -")
    def test_options(self, completion):
        assert completion

    @pytest.mark.complete("alias -p ")
    def test_p(self, completion):
        assert not completion



import pytest


@pytest.mark.bashcomp(pre_cmds=("INFOPATH+=:$PWD/info:",))
class TestInfo:
    @pytest.mark.complete("info bash")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("info -", require_cmd=True)
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete(
        "info nonexistent-na",
        env=dict(INFOPATH="'$(echo malicious code >/dev/tty)'"),
    )
    def test_infopath_code_injection(self, completion):
        # no completion, no space appended
        assert not completion



import pytest


class TestVgimport:
    @pytest.mark.complete(
        "vgimport -", require_cmd=True, xfail="! vgimport --help &>/dev/null"
    )
    def test_1(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(
    ignore_env=r"^-declare -f _comp_cmd_vncviewer__bootstrap$"
)
class TestVncviewer:
    @pytest.mark.complete("vncviewer ")
    def test_1(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(cmd="deja-dup")
class TestDejaDup:
    @pytest.mark.complete("deja-dup -", require_cmd=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("deja-dup --help ")
    def test_2(self, completion):
        assert not completion



import pytest

from conftest import assert_bash_exec


@pytest.mark.bashcomp(cmd=None, ignore_env=r"^[+-](___var|assoc[12])=")
class TestUnitCompgenVariables:
    @pytest.fixture(scope="class")
    def functions(self, request, bash):
        assert_bash_exec(bash, "unset assoc1 && declare -A assoc1=([idx]=1)")
        assert_bash_exec(
            bash, "unset assoc2 && declare -A assoc2=([idx1]=1 [idx2]=2)"
        )
        assert_bash_exec(bash, 'unset ${!___v*} && declare ___var=""')
        request.addfinalizer(
            lambda: assert_bash_exec(bash, "unset ___var assoc1 assoc2")
        )

    @pytest.mark.complete(": $___v")
    def test_simple_variable_name(self, functions, completion):
        assert completion == "ar"

    @pytest.mark.complete(": ${assoc1[")
    def test_single_array_index(self, functions, completion):
        assert completion == "idx]}"

    @pytest.mark.complete(": ${assoc2[")
    def test_multiple_array_indexes(self, functions, completion):
        assert completion == "${assoc2[idx1]} ${assoc2[idx2]}".split()

    @pytest.mark.complete(": ${assoc2[", shopt=dict(failglob=True))
    def test_multiple_array_indexes_failglob(self, functions, completion):
        assert completion == "${assoc2[idx1]} ${assoc2[idx2]}".split()

    @pytest.mark.complete(": ${assoc1[bogus]")
    def test_closing_curly_after_square(self, functions, completion):
        assert completion == "}"

    @pytest.mark.complete(": ${assoc1[@")
    def test_closing_brackets_after_at(self, functions, completion):
        assert completion == "]}"

    @pytest.mark.complete(": ${#___v")
    def test_hash_prefix(self, functions, completion):
        assert completion == "ar}"



import pytest


@pytest.mark.bashcomp(cmd="pylint-3")
class TestPylint3:
    @pytest.mark.complete("pylint-3 --v", require_cmd=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("pylint-3 http.clien")
    def test_2(self, completion):
        assert completion



import pytest


class TestSlabtop:
    @pytest.mark.complete("slabtop --", require_cmd=True)
    def test_basic(self, completion):
        assert completion

    @pytest.mark.complete("slabtop --sort ", require_cmd=True)
    def test_sort_arg(self, completion):
        assert completion



import pytest

from conftest import assert_bash_exec


class TestNmap:
    @pytest.fixture(scope="class")
    def functions(self, request, bash):
        assert_bash_exec(bash, "_mock_nmap() { cat nmap/nmap-h.txt; }")
        assert_bash_exec(bash, "complete -F _comp_cmd_nmap _mock_nmap")

    @pytest.mark.complete("nmap --v", require_cmd=True)
    def test_live_options(self, completion):
        assert completion

    @pytest.mark.complete("nmap ")
    def test_hosts(self, completion):
        assert completion

    @pytest.mark.complete("_mock_nmap -")
    def test_mock_options(self, completion, functions):
        assert completion == sorted(
            "-iL -iR --exclude --excludefile -sL -sn -Pn -PS -PA -PU -PY -PE "
            "-PP -PM -PO -n -R --dns-servers --system-dns --traceroute -sS "
            "-sT -sA -sW -sM -sU -sN -sF -sX --scanflags -sI -sY -sZ -sO -b "
            "-p --exclude-ports -F -r --top-ports --port-ratio -sV "
            "--version-intensity --version-light --version-all "
            "--version-trace -sC --script= --script-args= --script-args-file= "
            "--script-trace --script-updatedb --script-help= -O "
            "--osscan-limit --osscan-guess "
            # TODO: -T known mishandled; should expand -T<0-5> to -T0 ... -T5
            "-T --min-hostgroup --max-hostgroup --min-parallelism "
            "--max-parallelism --min-rtt-timeout --max-rtt-timeout "
            "--initial-rtt-timeout --max-retries --host-timeout --scan-delay "
            "--max-scan-delay --min-rate --max-rate -f --mtu -D -S -e -g "
            "--source-port --proxies --data --data-string --data-length "
            "--ip-options --ttl --spoof-mac --badsum -oN -oX -oS -oG -oA -v "
            "-d --reason --open --packet-trace --iflist --append-output "
            "--resume --stylesheet --webxml --no-stylesheet -6 -A --datadir "
            "--send-eth --send-ip --privileged --unprivileged -V -h"
            "".strip().split()
        )

    @pytest.mark.complete("_mock_nmap --script-args-f")
    def test_mock_nospace(self, completion, functions):
        assert completion == "ile="
        assert completion.endswith("=")  # no space appended



import pytest


@pytest.mark.bashcomp(cmd="qemu-system-x86_64")
class TestQemuSystemX8664:
    @pytest.mark.complete("qemu-system-x86_64 ")
    def test_basic(self, completion):
        assert completion

    @pytest.mark.complete("qemu-system-x86_64 -", require_cmd=True)
    def test_options(self, completion):
        assert completion

    @pytest.mark.complete("qemu-system-x86_64 -k ", require_cmd=True)
    def test_keymaps(self, completion):
        assert any(x.lower().startswith("en") for x in completion)



import pytest


class TestMock:
    @pytest.mark.complete("mock ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete(
        "mock -", require_cmd=True, xfail="! mock --help &>/dev/null"
    )
    def test_2(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(cmd="desktop-file-validate")
class TestDesktopFileValidate:
    @pytest.mark.complete("desktop-file-validate ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("desktop-file-validate -", require_cmd=True)
    def test_2(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(cmd="chromium-browser")
class TestChromiumBrowser:
    @pytest.mark.complete("chromium-browser ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete(
        "chromium-browser -", xfail="! chromium-browser --help &>/dev/null"
    )
    def test_2(self, completion):
        assert completion
        assert not completion.endswith(" ")

    @pytest.mark.complete("chromium-browser --proxy-server=")
    def test_proxy_server_scheme(self, completion):
        assert completion
        assert not completion.endswith(" ")
        assert all(x.endswith("://") for x in completion)

    @pytest.mark.complete("chromium-browser --proxy-server=http://")
    def test_proxy_server_host(self, completion):
        assert completion



from itertools import chain

import pytest


class TestSftp:
    @pytest.mark.complete("sftp -Fsp", cwd="sftp")
    def test_1(self, completion):
        assert completion == r"aced\ \ conf"

    @pytest.mark.complete("sftp -", require_cmd=True)
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("sftp -F config ", cwd="sftp")
    def test_hosts(self, hosts, completion):
        expected = sorted(
            chain(
                hosts,
                # From fixtures/sftp/config
                "gee hut".split(),
                # From fixtures/sftp/known_hosts
                "10.10.10.10 doo ike".split(),
            )
        )
        assert completion == expected

    @pytest.mark.complete(r"sftp -F spaced\ \ conf ", cwd="sftp")
    def test_hosts_spaced_conf(self, hosts, completion):
        expected = sorted(
            chain(
                hosts,
                # From "fixtures/sftp/spaced  conf"
                "gee jar".split(),
                # From fixtures/sftp/known_hosts
                "10.10.10.10 doo ike".split(),
            )
        )
        assert completion == expected

    @pytest.mark.complete("sftp -F")
    def test_capital_f_without_space(self, completion):
        assert completion
        assert not any(
            "option requires an argument -- F" in x for x in completion
        )



import pytest


@pytest.mark.bashcomp(ignore_env=r"^\+PERL5LIB=")
class TestPerl:
    @pytest.mark.complete("perl ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("perl -e ")
    def test_2(self, completion):
        assert not completion

    @pytest.mark.complete("perl -V:install", require_cmd=True)
    def test_3(self, completion):
        assert completion

    @pytest.mark.complete("perl -V::install", require_cmd=True)
    def test_4(self, completion):
        assert completion

    # Assume File::Spec and friends are always installed

    @pytest.mark.complete("perl -MFile", require_cmd=True)
    def test_5(self, completion):
        assert completion

    @pytest.mark.complete("perl -MFile::Sp", require_cmd=True)
    def test_6(self, completion):
        assert completion

    @pytest.mark.complete("perl -MFile::Spec::Func", require_cmd=True)
    def test_7(self, completion):
        assert completion

    @pytest.mark.complete("perl -M-File", require_cmd=True)
    def test_8(self, completion):
        assert completion

    @pytest.mark.complete("perl -m-File::", require_cmd=True)
    def test_9(self, completion):
        assert completion

    @pytest.mark.complete("perl -")
    def test_10(self, completion):
        assert completion

    @pytest.mark.complete("perl foo shared/default/f")
    def test_11(self, completion):
        """Second arg should complete files+dirs."""
        assert completion == "foo foo.d/".split()

    @pytest.mark.complete("perl -Ishared/default/")
    def test_12(self, completion):
        """-I without space should complete dirs."""
        assert completion == ["bar bar.d/", "foo.d/"]

    @pytest.mark.complete("perl -I shared/default/")
    def test_13(self, completion):
        """-I with space should complete dirs."""
        assert completion == ["bar bar.d/", "foo.d/"]

    @pytest.mark.complete("perl -xshared/default/b")
    def test_14(self, completion):
        """-x without space should complete dirs."""
        assert completion == r"ar\ bar.d/"

    @pytest.mark.complete("perl -x shared/default/b")
    def test_15(self, completion):
        """-x with space should complete files+dirs."""
        assert completion == ["bar", "bar bar.d/"]

    @pytest.mark.complete(
        "perl -d:", env=dict(PERL5LIB="$PWD/perl"), require_cmd=True
    )
    def test_16(self, completion):
        assert "BashCompletion" in completion

    @pytest.mark.complete(
        "perl -dt:", env=dict(PERL5LIB="$PWD/perl"), require_cmd=True
    )
    def test_17(self, completion):
        assert "BashCompletion" in completion

    @pytest.mark.complete("perl -E ")
    def test_dash_capital_e(self, completion):
        assert not completion

    @pytest.mark.complete("perl -e")
    def test_dash_e(self, completion):
        assert not completion



import pytest


class TestUserdel:
    @pytest.mark.complete("userdel -", require_cmd=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("userdel roo")
    def test_2(self, completion):
        assert completion == "t" or "root" in completion



import pytest


class TestIfstat:
    @pytest.mark.complete("ifstat -", require_cmd=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete(
        "ifstat -i ", xfail="ifstat -v | command grep -qF iproute2"
    )
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete(
        "ifstat -d ",
        require_cmd=True,
        xfail="ifstat -v | command grep -qF iproute2",
    )
    def test_3(self, completion):
        assert completion



import os

import pytest


class TestRpm2tgz:
    @pytest.mark.complete("rpm2tgz -")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("rpm2tgz ", cwd="slackware/home")
    def test_2(self, completion):
        expected = sorted(
            [
                "%s/" % x
                for x in os.listdir("slackware/home")
                if os.path.isdir("slackware/home/%s" % x)
            ]
            + [
                x
                for x in os.listdir("slackware/home")
                if os.path.isfile("slackware/home/%s" % x)
                and x.endswith(".rpm")
            ]
        )
        assert completion == expected



import pytest


@pytest.mark.bashcomp(
    pre_cmds=("HOME=$PWD/cvs",),
    ignore_env=r"^[+-]COMP_CVS_REMOTE=",
)
class TestCvsps:
    @pytest.mark.complete("cvsps -", require_cmd=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("cvsps ")
    def test_2(self, completion):
        assert [x for x in completion if ":pserver:" in x]



import getpass

import pytest


class TestPostfix:
    @pytest.mark.complete("postfix ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.xfail(
        getpass.getuser() != "root",
        reason="Likely outputs usage only for root",
    )
    @pytest.mark.complete(
        "postfix -",
        require_cmd=True,
        xfail="! type unbuffer &>/dev/null",
        sleep_after_tab=2,  # postfix is slow to output usage
    )
    def test_options(self, completion):
        assert completion



import pytest

from conftest import assert_bash_exec, in_container


@pytest.mark.bashcomp(cmd=None, ignore_env=r"^\+COMPREPLY=")
class TestUnitCompgenIpAddresses:
    @pytest.fixture(scope="class")
    def functions(self, request, bash):
        assert_bash_exec(
            bash,
            "_ia() { local cur;_comp_get_words cur;"
            "unset -v COMPREPLY;_comp_compgen_ip_addresses; }",
        )
        assert_bash_exec(bash, "complete -F _ia ia")
        assert_bash_exec(
            bash,
            "_iaa() { local cur;_comp_get_words cur;"
            "unset -v COMPREPLY;_comp_compgen_ip_addresses -a; }",
        )
        assert_bash_exec(bash, "complete -F _iaa iaa")
        assert_bash_exec(
            bash,
            " _ia6() { local cur;_comp_get_words cur;"
            "unset -v COMPREPLY;_comp_compgen_ip_addresses -6; }",
        )
        assert_bash_exec(bash, "complete -F _ia6 ia6")

    def test_1(self, bash):
        assert_bash_exec(bash, "_comp_compgen_ip_addresses")

    @pytest.mark.complete("iaa ")
    def test_2(self, functions, completion):
        """_comp_compgen_ip_addresses -a should complete ip addresses."""
        assert completion
        assert all("." in x or ":" in x for x in completion)

    @pytest.mark.complete("ia ")
    def test_3(self, functions, completion):
        """_comp_compgen_ip_addresses should complete ipv4 addresses."""
        assert completion
        assert all("." in x for x in completion)

    @pytest.mark.xfail(in_container(), reason="Probably fails in a container")
    @pytest.mark.complete("ia6 ")
    def test_4(self, functions, completion):
        """_comp_compgen_ip_addresses -6 should complete ipv6 addresses."""
        assert completion
        assert all(":" in x for x in completion)



import pytest


@pytest.mark.bashcomp(
    ignore_env=r"^-declare -f _comp_cmd_makepkg__bootstrap$",
    xfail="! makepkg --help 2>&1 | command grep -qiF slackware",
)
class TestMakepkg:
    @pytest.mark.complete("makepkg ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("makepkg --", require_cmd=True)
    def test_2(self, completion):
        assert all(
            x in completion for x in "--chown --linkadd --prepend".split()
        )



import pytest


class TestPerltidy:
    @pytest.mark.complete("perltidy ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("perltidy -h", require_cmd=True)
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("perltidy -ole=", require_cmd=True)
    def test_3(self, completion):
        assert completion

    @pytest.mark.complete("perltidy -doesntexist=", require_cmd=True)
    def test_4(self, completion):
        assert not completion



import pytest


class TestChroot:
    @pytest.mark.complete("chroot ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("/bin/chroot shared/default/")
    def test_2(self, completion):
        """Should complete dirs only, also when invoked using full path."""
        assert completion == ["bar bar.d/", "foo.d/"]

    @pytest.mark.complete("chroot -", require_longopt=True)
    def test_options(self, completion):
        assert completion



import pytest


class TestVgremove:
    @pytest.mark.complete(
        "vgremove -", require_cmd=True, xfail="! vgremove --help &>/dev/null"
    )
    def test_1(self, completion):
        assert completion



import pytest


class TestPycodestyle:
    @pytest.mark.complete("pycodestyle ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("pycodestyle -", require_cmd=True)
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("pycodestyle --doesnt-exist=")
    def test_3(self, completion):
        assert not completion



import pytest


@pytest.mark.bashcomp(cmd="xdg-mime")
class TestXdgMime:
    @pytest.mark.complete("xdg-mime ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("xdg-mime -")
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("xdg-mime query ")
    def test_3(self, completion):
        assert completion

    @pytest.mark.complete("xdg-mime query filetype ")
    def test_4(self, completion):
        assert completion

    @pytest.mark.complete("xdg-mime default foo.desktop ", require_cmd=True)
    def test_5(self, completion):
        assert completion

    @pytest.mark.complete("xdg-mime install --mode ")
    def test_6(self, completion):
        assert completion

    @pytest.mark.complete("xdg-mime query filetype foo ")
    def test_filetype_one_arg(self, completion):
        assert not completion



import os

import pytest


@pytest.mark.bashcomp(ignore_env=r"^\+ROOT=")
class TestRemovepkg:
    @pytest.mark.complete("removepkg -")
    def test_1(self, completion):
        assert completion == "-copy -keep -preserve -warn".split()

    @pytest.mark.complete("removepkg ", env=dict(ROOT="./slackware"))
    def test_2(self, completion):
        files = sorted(x for x in os.listdir("slackware/var/log/packages"))
        assert completion == files



import pytest

from conftest import assert_bash_exec, bash_env_saved


@pytest.mark.bashcomp(
    cmd=None, cwd="shared", ignore_env=r"^\+declare -f __tester$"
)
class TestUnitRealCommand:
    @pytest.fixture
    def functions(self, bash):
        assert_bash_exec(
            bash,
            (
                "__tester() { "
                "local REPLY rc; "
                '_comp_realcommand "$1"; '
                "rc=$?; "
                'printf %s "$REPLY"; '
                "return $rc; "
                "}"
            ),
        )

    def test_non_pollution(self, bash):
        """Test environment non-pollution, detected at teardown."""
        assert_bash_exec(
            bash,
            "foo() { local REPLY=; _comp_realcommand bar; }; foo; unset -f foo",
            want_output=None,
        )

    def test_basename(self, bash, functions):
        with bash_env_saved(bash) as bash_env:
            bash_env.write_variable("PATH", "$PWD/bin:$PATH", quote=False)
            output = assert_bash_exec(
                bash,
                "__tester arp",
                want_output=True,
                want_newline=False,
            )
            assert output.strip().endswith("/shared/bin/arp")

    def test_basename_nonexistent(self, bash, functions):
        filename = "non-existent-file-for-bash-completion-tests"
        file_exists_check = "! type -P %s" % filename
        try:
            assert_bash_exec(bash, file_exists_check, want_output=None)
        except AssertionError:
            pytest.skip(f"Found command that shouldn't exist: '{filename}'")
        output = assert_bash_exec(
            bash,
            "! __tester %s" % filename,
            want_output=False,
        )
        assert output.strip() == ""

    def test_relative(self, bash, functions):
        output = assert_bash_exec(
            bash,
            "__tester bin/arp",
            want_output=True,
            want_newline=False,
        )
        assert output.strip().endswith("/shared/bin/arp")

    def test_relative_nonexistent(self, bash, functions):
        output = assert_bash_exec(
            bash,
            "! __tester bin/non-existent",
            want_output=False,
        )
        assert output.strip() == ""

    def test_absolute(self, bash, functions):
        output = assert_bash_exec(
            bash,
            '__tester "$PWD/bin/arp"',
            want_output=True,
            want_newline=False,
        )
        assert output.strip().endswith("/shared/bin/arp")

    def test_absolute_nonexistent(self, bash, functions):
        output = assert_bash_exec(
            bash,
            '! __tester "$PWD/bin/non-existent"',
            want_output=False,
        )
        assert output.strip() == ""

    def test_option_like_cmd_name(self, bash, functions):
        output = assert_bash_exec(
            bash,
            "! __tester --non-existent",
            want_output=False,
        )
        assert output.strip() == ""



import os

import pytest


@pytest.mark.bashcomp(pre_cmds=("PKG_DBDIR=$PWD/pkgtools/db",))
class TestPkgDeinstall:
    @pytest.mark.complete("pkg_deinstall ")
    def test_1(self, completion):
        dirs = sorted(
            x
            for x in os.listdir("pkgtools/db")
            if os.path.isdir("pkgtools/db/%s" % x)
        )
        assert completion == dirs



import pytest


class TestSysctl:
    @pytest.mark.complete("sysctl -", require_cmd=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete(
        "sysctl kern",
        require_cmd=True,
        xfail="! sysctl -N -a 2>/dev/null | command grep -q ^kern",
    )
    def test_2(self, completion):
        assert completion



import pytest


class TestLsusb:
    @pytest.mark.complete(
        "lsusb -",
        xfail="! (lsusb --help 2>&1 || :) | command grep -qF -- --help",
    )
    def test_1(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(cmd="gssdp-discover")
class TestGssdpDiscover:
    @pytest.mark.complete("gssdp-discover ")
    def test_no_args(self, completion):
        assert not completion

    @pytest.mark.complete("gssdp-discover --", require_cmd=True)
    def test_options(self, completion):
        assert completion

    @pytest.mark.complete(
        "gssdp-discover --message-type=",
        skipif=(
            "! gssdp-discover --help 2>&1 "
            "| command grep -qF -- --message-type"
        ),
    )
    def test_message_type(self, completion):
        assert completion



import pytest


class TestInflux:
    @pytest.mark.complete("influx ")
    def test_nothing(self, completion):
        assert not completion

    @pytest.mark.complete("influx -", require_cmd=True)
    def test_options(self, completion):
        assert completion

    @pytest.mark.complete("influx -format ", require_cmd=True)
    def test_format(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(pre_cmds=("HOME=$PWD/lftp",))
class TestLftp:
    @pytest.mark.complete("lftp ", require_cmd=True)
    def test_1(self, bash, completion, output_sort_uniq):
        hosts = output_sort_uniq("compgen -A hostname")
        assert all(x in completion for x in hosts)
        # defined in lftp/.lftp/bookmarks
        assert all(x in completion for x in "lftptest spacetest".split())
        assert "badbookmark" not in completion

    @pytest.mark.complete("lftp -", require_cmd=True)
    def test_2(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(pre_cmds=("PKG_DBDIR=$PWD/pkgtools/db",))
class TestPortupgrade:
    @pytest.mark.complete("portupgrade ")
    def test_1(self, completion):
        assert completion == "a b-c-d".split()



import pytest

from conftest import assert_bash_exec


@pytest.mark.bashcomp(pre_cmds=("HOME=$PWD/mutt",))
class TestMutt:
    @pytest.fixture(scope="class")
    def functions(self, bash):
        assert_bash_exec(
            bash,
            '_comp_test__muttconffiles() { local REPLY; _comp_cmd_mutt__get_conffiles "$@" && printf "%s\\n" "${REPLY[@]}"; }',
        )

    @pytest.mark.complete("mutt -")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("mutt -F muttrc -f =", require_cmd=True, cwd="mutt")
    def test_2(self, completion):
        assert completion == "bar/ foo/ muttrc".split()

    @pytest.mark.complete("mutt -F muttrc -A ", cwd="mutt")
    def test_3(self, completion):
        assert completion == "a1 a2".split()

    def test_4(self, bash, functions):
        got = (
            assert_bash_exec(
                bash,
                '_comp_test__muttconffiles "$HOME/muttrc" "$HOME/muttrc"',
                want_output=True,
            )
            .strip()
            .split()
        )
        assert got == [
            f"{bash.cwd}/mutt/{x}"
            for x in ("muttrc", "bar/muttrc_b", "foo/muttrc_f")
        ]



import pytest


class TestPidof:
    # "p": Assume that our process name completion runs ps
    @pytest.mark.complete("pidof p")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete(
        "pidof -", require_cmd=True, xfail="! pidof --help &>/dev/null"
    )
    def test_2(self, completion):
        assert completion



import pytest


class TestSynclient:
    # synclient -l may error out with e.g.
    # "Couldn't find synaptics properties. No synaptics driver loaded?"
    @pytest.mark.complete(
        "synclient ", require_cmd=True, xfail="! synclient -l &>/dev/null"
    )
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("synclient -", require_cmd=True)
    def test_2(self, completion):
        assert completion



import pytest


class TestAptitude:
    @pytest.mark.complete("aptitude ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("aptitude -", require_cmd=True)
    def test_options(self, completion):
        assert completion

    @pytest.mark.complete("aptitude --", require_cmd=True)
    def test_long_options(self, completion):
        assert completion

    @pytest.mark.complete("aptitude -u -")
    def test_no_i_with_u(self, completion):
        assert "-i" not in completion

    @pytest.mark.complete("aptitude -i -")
    def test_no_u_with_i(self, completion):
        assert "-u" not in completion



import pytest


class TestArpspoof:
    @pytest.mark.complete(
        "arpspoof -",
        require_cmd=True,
        # May require privileges or network interfaces available even for
        # outputting the usage message. Unfortunately --help provokes a
        # non-zero exit status so we cannot test for that.
        skipif=(
            "arpspoof 2>&1 | "
            "command grep -qE 'libnet_(open_link|select_device)'"
        ),
    )
    def test_1(self, completion):
        assert completion



import pytest


class TestFirefox:
    @pytest.mark.complete("firefox ")
    def test_1(self, completion):
        assert completion

    # --help test: running as root in GH actions container croaks:
    #     Running Firefox as root in a regular user's session is not supported.
    #     ($HOME is /github/home which is owned by uid 1001.)
    @pytest.mark.complete(
        "firefox -", require_cmd=True, xfail="! firefox --help &>/dev/null"
    )
    def test_2(self, completion):
        assert completion
        assert not completion.endswith(" ")



import pytest


class TestPvchange:
    @pytest.mark.complete(
        "pvchange --", require_cmd=True, xfail="! pvchange --help &>/dev/null"
    )
    def test_1(self, completion):
        assert completion



import pytest


class Test7z:
    @pytest.mark.complete("7z ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("7z a ar -tzi")
    def test_2(self, completion):
        assert completion == "p"

    @pytest.mark.complete(r"7z x -wa\ ", cwd="_filedir")
    def test_3(self, completion):
        assert completion == "b/"
        assert not completion.endswith(" ")

    @pytest.mark.complete("7z x ", cwd="7z")
    def test_4(self, completion):
        assert completion == "a.7z hello.7z.001".split()

    @pytest.mark.complete("7z d a.7z ", cwd="7z", require_cmd=True)
    def test_5(self, completion):
        assert completion == "abc"

    @pytest.mark.complete("7z a -air@", cwd="7z")
    def test_6(self, completion):
        assert completion == sorted(
            "-air@a.7z -air@hello.7z.001 -air@hello.7z.002 -air@f.txt".split()
        )

    @pytest.mark.complete("7z a -o")
    def test_7(self, completion):
        assert "-o7z/" in completion
        assert all(x.endswith("/") for x in completion)



import pytest


class TestGpgv:
    @pytest.mark.complete("gpgv ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("gpgv -", require_cmd=True)
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("gpgv foo.sig foo ")
    def test_3(self, completion):
        assert not completion



import pytest


class TestSha256sum:
    @pytest.mark.complete("sha256sum --", require_longopt=True)
    def test_options(self, completion):
        assert completion

    @pytest.mark.complete("sha256sum ", cwd="sha256sum")
    def test_summing(self, completion):
        assert completion == "foo"

    @pytest.mark.complete("sha256sum -c ", cwd="sha256sum")
    def test_checking(self, completion):
        assert completion == "foo.sha256"



import os

import pytest

from conftest import assert_bash_exec, assert_complete, bash_env_saved


@pytest.mark.bashcomp(cmd=None, temp_cwd=True)
class TestUnitQuoteCompgen:
    @pytest.fixture(scope="class")
    def functions(self, bash):
        assert_bash_exec(
            bash,
            '_comp__test_quote_compgen() { local REPLY; _comp_quote_compgen "$1"; printf %s "$REPLY"; }',
        )

    @pytest.mark.parametrize(
        "funcname", "_comp__test_quote_compgen quote_readline".split()
    )
    def test_exec(self, bash, functions, funcname):
        assert_bash_exec(bash, "%s '' >/dev/null" % funcname)

    @pytest.mark.parametrize(
        "funcname", "_comp__test_quote_compgen quote_readline".split()
    )
    def test_env_non_pollution(self, bash, functions, funcname):
        """Test environment non-pollution, detected at teardown."""
        assert_bash_exec(
            bash, "foo() { %s meh >/dev/null; }; foo; unset -f foo" % funcname
        )

    @pytest.mark.parametrize(
        "funcname", "_comp__test_quote_compgen quote_readline".split()
    )
    def test_1(self, bash, functions, funcname):
        output = assert_bash_exec(
            bash, "%s '';echo" % funcname, want_output=True
        )
        assert output.strip() == "''"

    @pytest.mark.parametrize(
        "funcname", "_comp__test_quote_compgen quote_readline".split()
    )
    def test_2(self, bash, functions, funcname):
        output = assert_bash_exec(
            bash, "%s foo;echo" % funcname, want_output=True
        )
        assert output.strip() == "foo"

    @pytest.mark.parametrize(
        "funcname", "_comp__test_quote_compgen quote_readline".split()
    )
    def test_3(self, bash, functions, funcname):
        output = assert_bash_exec(
            bash, '%s foo\\"bar;echo' % funcname, want_output=True
        )
        assert output.strip() == 'foo\\"bar'

    @pytest.mark.parametrize(
        "funcname", "_comp__test_quote_compgen quote_readline".split()
    )
    def test_4(self, bash, functions, funcname):
        output = assert_bash_exec(
            bash, "%s '$(echo x >&2)';echo" % funcname, want_output=True
        )
        assert output.strip() == "\\$\\(echo\\ x\\ \\>\\&2\\)"

    def test_github_issue_189_1(self, bash, functions):
        """Test error messages on a certain command line

        Reported at https://github.com/scop/bash-completion/issues/189

        Syntax error messages should not be shown by completion on the
        following line:

          $ ls -- '${[TAB]
          $ rm -- '${[TAB]

        """
        assert_bash_exec(bash, "_comp__test_quote_compgen $'\\'${' >/dev/null")

    def test_github_issue_492_1(self, bash, functions):
        """Test unintended code execution on a certain command line

        Reported at https://github.com/scop/bash-completion/pull/492

        Arbitrary commands could be unintendedly executed by
        _comp_quote_compgen.  In the following example, the command "touch
        1.txt" would be unintendedly created before the fix.  The file "1.txt"
        should not be created by completion on the following line:

          $ echo '$(touch file.txt)[TAB]

        """
        assert_bash_exec(
            bash, "_comp__test_quote_compgen $'\\'$(touch 1.txt)' >/dev/null"
        )
        assert not os.path.exists("./1.txt")

    def test_github_issue_492_2(self, bash, functions):
        """Test the file clear by unintended redirection on a certain command line

        Reported at https://github.com/scop/bash-completion/pull/492

        The file "1.0" should not be created by completion on the following
        line:

          $ awk '$1 > 1.0[TAB]

        """
        assert_bash_exec(
            bash, "_comp__test_quote_compgen $'\\'$1 > 1.0' >/dev/null"
        )
        assert not os.path.exists("./1.0")

    def test_github_issue_492_3(self, bash, functions):
        """Test code execution through unintended pathname expansions

        When there is a file named "quote=$(COMMAND)" (for
        _comp_compgen_filedir) or "REPLY=$(COMMAND)" (for _comp_quote_compgen),
        the completion of the word '$* results in the execution of COMMAND.

          $ echo '$*[TAB]

        """
        os.mkdir("./REPLY=$(echo injected >&2)")
        assert_bash_exec(bash, "_comp__test_quote_compgen $'\\'$*' >/dev/null")

    def test_github_issue_492_4(self, bash, functions):
        """Test error messages through unintended pathname expansions

        When "shopt -s failglob" is set by the user, the completion of the word
        containing glob character and special characters (e.g. TAB) results in
        the failure of pathname expansions.

          $ shopt -s failglob
          $ echo a\\	b*[TAB]

        """
        with bash_env_saved(bash) as bash_env:
            bash_env.shopt("failglob", True)
            assert_bash_exec(
                bash, "_comp__test_quote_compgen $'a\\\\\\tb*' >/dev/null"
            )

    def test_github_issue_526_1(self, bash):
        r"""Regression tests for unprocessed escape sequences after quotes

        Ref [1] https://github.com/scop/bash-completion/pull/492#discussion_r637213822
        Ref [2] https://github.com/scop/bash-completion/pull/526

        The escape sequences in the local variable of "value" in
        "_comp_quote_compgen" needs to be unescaped by passing it to printf as
        the format string.  This causes a problem in the following case [where
        the spaces after "alpha\" is a TAB character inserted in the command
        string by "C-v TAB"]:

          $ echo alpha\   b[TAB]

        """
        os.mkdir("./alpha\tbeta")
        assert (
            assert_complete(
                # Remark on "rendered_cmd": Bash aligns the last character 'b'
                # in the rendered cmd to an "8 x n" boundary using spaces.
                # Here, the command string is assumed to start from column 2
                # because the width of PS1 (conftest.PS1 = '/@') is 2,
                bash,
                "echo alpha\\\026\tb",
                rendered_cmd="echo alpha\\   b",
            )
            == "eta/"
        )



import pytest


class TestUsermod:
    @pytest.mark.complete("usermod ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("usermod -", require_cmd=True)
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("useradd --root shells -s ")
    def test_chroot_shells(self, completion):
        assert completion == "/bash/completion/canary"



import pytest

from conftest import assert_bash_exec


@pytest.mark.bashcomp(cmd=None)
class TestUnitDelimited:
    @pytest.fixture(scope="class")
    def functions(self, request, bash):
        assert_bash_exec(
            bash,
            "_comp_cmd_test_delim() {"
            "  local cur prev words cword comp_args;"
            "  _comp_get_words cur;"
            "  _comp_delimited , -W 'alpha beta bravo';"
            "};"
            "complete -F _comp_cmd_test_delim test_delim",
        )

    @pytest.mark.complete("test_delim --opt=a")
    def test_1(self, functions, completion):
        assert completion == ["lpha"]

    @pytest.mark.complete("test_delim --opt=b")
    def test_2(self, functions, completion):
        assert completion == ["beta", "bravo"]

    @pytest.mark.complete("test_delim --opt=alpha,b")
    def test_3(self, functions, completion):
        assert completion == ["alpha,beta", "alpha,bravo"]

    @pytest.mark.complete("test_delim --opt=alpha,be")
    def test_4(self, functions, completion):
        assert completion == ["ta"]

    @pytest.mark.complete("test_delim --opt=beta,a")
    def test_5(self, functions, completion):
        assert completion == ["lpha"]

    @pytest.mark.complete("test_delim --opt=c")
    def test_6(self, functions, completion):
        assert not completion



import pytest


class TestGdb:
    @pytest.mark.complete("gdb - ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("gdb foo ", cwd="gdb")
    def test_2(self, completion):
        assert completion == sorted(
            "core core.12345 "
            "core.weston.1000.deadbeef.5308.1555362132000000".split()
        )

    @pytest.mark.complete("gdb aw")
    def test_3(self, completion):
        """Check that the completion can generate command names"""
        assert completion == ["k"] or "awk" in completion

    @pytest.mark.complete("gdb built")
    def test_4(self, completion):
        """Check that the completion does not generate builtin names"""
        assert not (completion == ["in"] or "builtin" in completion)



import pytest


class TestTree:
    @pytest.mark.complete("tree ", cwd="shared/default")
    def test_basic(self, completion):
        assert completion == ["bar bar.d/", "foo.d/"]

    @pytest.mark.complete("tree --fromfile ", cwd="shared/default")
    def test_fromfile(self, completion):
        assert completion == ["bar", "bar bar.d/", "foo", "foo.d/"]

    @pytest.mark.complete(
        "tree -",
        require_cmd=True,
        xfail="! tree --help 2>&1 | command grep -qF -- ' -'",
    )
    def test_options(self, completion):
        assert completion

    @pytest.mark.complete("tree --sort=", require_cmd=True)
    def test_equals_sign_split(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(cmd="btdownloadheadless.py")
class TestBtdownloadheadlessPy:
    @pytest.mark.complete("btdownloadheadless.py ")
    def test_1(self, completion):
        assert completion



import pytest


class TestLvremove:
    @pytest.mark.complete(
        "lvremove --", require_cmd=True, xfail="! lvremove --help &>/dev/null"
    )
    def test_1(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(
    cmd="fprintd-delete",
)
class TestFprintdDelete:
    @pytest.mark.complete("fprintd-delete ")
    def test_basic(self, completion):
        assert completion



import pytest


class TestPgrep:
    # "p": Assume that our process name completion runs ps
    @pytest.mark.complete("pgrep p")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("pgrep -", require_cmd=True)
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete(
        "pgrep --nslist ",
        require_cmd=True,
        skipif=(
            "! pgrep --help 2>&1 | command grep -qF 'Available namespaces'"
        ),
    )
    def test_nslist(self, completion):
        assert completion
        assert not any("," in x for x in completion)

    @pytest.mark.complete(
        "pgrep --nslist foo,",
        require_cmd=True,
        skipif=(
            "! pgrep --help 2>&1 | command grep -qF 'Available namespaces'"
        ),
    )
    def test_nslist_after_comma(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(
    cmd="fprintd-verify",
)
class TestFprintdVerify:
    @pytest.mark.complete("fprintd-verify ")
    def test_basic(self, completion):
        assert completion

    @pytest.mark.complete("fprintd-verify -", require_cmd=True)
    def test_options(self, completion):
        assert completion



import pytest


class TestLspci:
    @pytest.mark.complete("lspci -", require_cmd=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete(
        "lspci -A ", require_cmd=True, skipif="! lspci -A help &>/dev/null"
    )
    def test_2(self, completion):
        assert completion



import pytest


@pytest.mark.bashcomp(cmd="mii-diag")
class TestMiiDiag:
    @pytest.mark.complete("mii-diag ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("mii-diag -", require_cmd=True)
    def test_2(self, completion):
        assert completion



import pytest

from conftest import assert_bash_exec, assert_complete


@pytest.mark.bashcomp(ignore_env=r"^[+-]_comp_cmd_scp__path_esc=")
class TestRsync:
    @pytest.mark.complete("rsync ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("rsync --rsh ")
    def test_2(self, completion):
        assert completion == "rsh ssh".split()

    @pytest.mark.complete("rsync --rsh=")
    def test_3(self, completion):
        assert completion == "rsh ssh".split()

    @pytest.mark.complete("rsync --", require_cmd=True)
    def test_4(self, completion):
        assert "--help" in completion

    @pytest.mark.parametrize(
        "ver1,ver2,result",
        [
            ("1", "1", "="),
            ("1", "2", "<"),
            ("2", "1", ">"),
            ("1.1", "1.2", "<"),
            ("1.2", "1.1", ">"),
            ("1.1", "1.1.1", "<"),
            ("1.1.1", "1.1", ">"),
            ("1.1.1", "1.1.1", "="),
            ("2.1", "2.2", "<"),
            ("3.0.4.10", "3.0.4.2", ">"),
            ("4.08", "4.08.01", "<"),
            ("3.2.1.9.8144", "3.2", ">"),
            ("3.2", "3.2.1.9.8144", "<"),
            ("1.2", "2.1", "<"),
            ("2.1", "1.2", ">"),
            ("5.6.7", "5.6.7", "="),
            ("1.01.1", "1.1.1", "="),
            ("1.1.1", "1.01.1", "="),
            ("1", "1.0", "="),
            ("1.0", "1", "="),
            ("1.0.2.0", "1.0.2", "="),
            ("1..0", "1.0", "="),
            ("1.0", "1..0", "="),
        ],
    )
    def test_vercomp(self, bash, ver1, ver2, result):
        output = assert_bash_exec(
            bash,
            f"_comp_cmd_rsync__vercomp {ver1} {ver2}; echo $?",
            want_output=True,
        ).strip()

        if result == "=":
            assert output == "0"
        elif result == ">":
            assert output == "1"
        elif result == "<":
            assert output == "2"
        else:
            raise Exception(f"Unsupported comparison result: {result}")

    def test_remote_path_with_spaces(self, bash):
        assert_bash_exec(bash, "ssh() { echo 'spaces in filename.txt'; }")
        completion = assert_complete(bash, "rsync remote_host:spaces")
        assert_bash_exec(bash, "unset -f ssh")
        assert (
            completion == r"\ in\ filename.txt"
            or completion == r"\\\ in\\\ filename.txt"
        )



import pytest

from conftest import TestUnitBase, assert_bash_exec, assert_complete


@pytest.mark.bashcomp(
    cmd=None,
    ignore_env=r"^[+-](COMP(_(WORDS|CWORD|LINE|POINT)|REPLY)|"
    r"cur|prev|cword|words)=|^\+declare -f _cmd1$",
)
class TestUnitInitialize(TestUnitBase):
    def test_1(self, bash):
        """Test environment non-pollution, detected at teardown."""
        assert_bash_exec(
            bash,
            "foo() { "
            "local cur prev words cword comp_args "
            "COMP_WORDS=() COMP_CWORD=0 COMP_LINE= COMP_POINT=0; "
            "_comp_initialize; }; "
            "foo; unset -f foo",
        )

    def test_2(self, bash):
        output = self._test_unit(
            "_comp_initialize %s; echo $cur,${prev-}", bash, "(a)", 0, "a", 0
        )
        assert output == ","

    @pytest.mark.parametrize("redirect", "> >> 2> < &>".split())
    def test_redirect(self, bash, redirect):
        completion = assert_complete(
            bash, "%s " % redirect, cwd="shared/default"
        )
        assert all(x in completion for x in "foo bar".split())

    @pytest.fixture(scope="class")
    def cmd1_empty_completion_setup(self, bash):
        assert_bash_exec(
            bash,
            '_cmd1() { local cur prev words cword comp_args; _comp_initialize -- "$@"; } && '
            "complete -F _cmd1 cmd1",
        )

    @pytest.mark.parametrize("redirect", "> >> 2> {fd1}> < &> &>> >|".split())
    def test_redirect_2(self, bash, cmd1_empty_completion_setup, redirect):
        # Note: Bash 4.3 and below cannot properly extract the redirection ">|"
        if redirect == ">|":
            skipif = "((BASH_VERSINFO[0] * 100 + BASH_VERSINFO[1] < 404))"
            try:
                assert_bash_exec(bash, skipif, want_output=None)
            except AssertionError:
                pass
            else:
                pytest.skip(skipif)

        completion = assert_complete(
            bash, "cmd1 %s f" % redirect, cwd="shared/default"
        )
        assert "foo" in completion

    @pytest.mark.parametrize("redirect", "> >> 2> < &>".split())
    def test_redirect_3(self, bash, redirect):
        completion = assert_complete(
            bash, "cmd1 %sf" % redirect, cwd="shared/default"
        )
        assert "foo" in completion



import pytest


class TestReportbug:
    @pytest.mark.complete("reportbug --m", require_cmd=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("reportbug --bts=", require_cmd=True)
    def test_bts(self, completion):
        assert "default" in completion



import pytest


class TestPython3:
    @pytest.mark.complete("python3 ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("python3 -", require_cmd=True)
    def test_2(self, completion):
        assert len(completion) > 1

    @pytest.mark.complete("python3 -c ")
    def test_3(self, completion):
        assert not completion

    @pytest.mark.complete("python3 shared/default/")
    def test_4(self, completion):
        assert completion == ["bar bar.d/", "foo.d/"]

    @pytest.mark.complete("python3 -c foo shared/default/")
    def test_5(self, completion):
        assert completion == ["bar", "bar bar.d/", "foo", "foo.d/"]

    @pytest.mark.complete("python3 -c foo -")
    def test_6(self, completion):
        assert not completion

    @pytest.mark.complete("python3 -m foo -")
    def test_7(self, completion):
        assert not completion

    @pytest.mark.complete("python3 -m sy", require_cmd=True)
    def test_8(self, completion):
        assert completion

    @pytest.mark.complete(
        "python3 -b",
        require_cmd=True,
        skipif="! python3 -h | command grep -qwF -- -bb",
    )
    def test_bb(self, completion):
        assert "-bb" in completion



import os

import pytest


@pytest.mark.xfail(not os.environ.get("DISPLAY"), reason="X display required")
class TestGkrellm:
    @pytest.mark.complete("gkrellm -", require_cmd=True)
    def test_1(self, completion):
        assert completion



import pytest

from conftest import assert_bash_exec, bash_env_saved


@pytest.mark.bashcomp(
    cmd=None,
    cwd="_filedir",
    ignore_env=r"^\+declare -f __tester$",
)
class TestDequote:
    def test_1_char(self, bash):
        assert_bash_exec(
            bash,
            '__tester() { local REPLY=dummy v=var;_comp_dequote "$1";local ext=$?;((${#REPLY[@]}))&&printf \'<%s>\' "${REPLY[@]}";echo;return $ext;}',
        )
        output = assert_bash_exec(bash, "__tester a", want_output=True)
        assert output.strip() == "<a>"

    def test_2_str(self, bash):
        output = assert_bash_exec(bash, "__tester abc", want_output=True)
        assert output.strip() == "<abc>"

    def test_3_null(self, bash):
        output = assert_bash_exec(bash, "__tester ''", want_output=True)
        assert output.strip() == ""

    def test_4_empty(self, bash):
        output = assert_bash_exec(bash, "__tester \"''\"", want_output=True)
        assert output.strip() == "<>"

    def test_5_brace(self, bash):
        output = assert_bash_exec(bash, "__tester 'a{1..3}'", want_output=True)
        assert output.strip() == "<a1><a2><a3>"

    def test_6_glob(self, bash):
        output = assert_bash_exec(bash, "__tester 'a?b'", want_output=True)
        assert output.strip() == "<a b><a$b><a&b><a'b>"

    def test_7_quote_1(self, bash):
        output = assert_bash_exec(
            bash, "__tester '\"a\"'\\'b\\'\\$\\'c\\'", want_output=True
        )
        assert output.strip() == "<abc>"

    def test_7_quote_2(self, bash):
        output = assert_bash_exec(
            bash, "__tester '\\\"\\'\\''\\$\\`'", want_output=True
        )
        assert output.strip() == "<\"'$`>"

    def test_7_quote_3(self, bash):
        output = assert_bash_exec(
            bash, "__tester \\$\\'a\\\\tb\\'", want_output=True
        )
        assert output.strip() == "<a\tb>"

    def test_7_quote_4(self, bash):
        output = assert_bash_exec(
            bash, '__tester \'"abc\\"def"\'', want_output=True
        )
        assert output.strip() == '<abc"def>'

    def test_7_quote_5(self, bash):
        output = assert_bash_exec(
            bash, "__tester \\'abc\\'\\\\\\'\\'def\\'", want_output=True
        )
        assert output.strip() == "<abc'def>"

    def test_8_param_1(self, bash):
        output = assert_bash_exec(bash, "__tester '$v'", want_output=True)
        assert output.strip() == "<var>"

    def test_8_param_2(self, bash):
        output = assert_bash_exec(bash, "__tester '${v}'", want_output=True)
        assert output.strip() == "<var>"

    def test_8_param_3(self, bash):
        output = assert_bash_exec(bash, "__tester '${#v}'", want_output=True)
        assert output.strip() == "<3>"

    def test_8_param_4(self, bash):
        output = assert_bash_exec(bash, "__tester '${v[0]}'", want_output=True)
        assert output.strip() == "<var>"

    def test_9_qparam_1(self, bash):
        output = assert_bash_exec(bash, "__tester '\"$v\"'", want_output=True)
        assert output.strip() == "<var>"

    def test_9_qparam_2(self, bash):
        output = assert_bash_exec(
            bash, "__tester '\"${v[@]}\"'", want_output=True
        )
        assert output.strip() == "<var>"

    def test_10_pparam_1(self, bash):
        output = assert_bash_exec(bash, "__tester '$?'", want_output=True)
        assert output.strip() == "<0>"

    def test_10_pparam_2(self, bash):
        output = assert_bash_exec(bash, "__tester '${#1}'", want_output=True)
        assert output.strip() == "<5>"  # The string `${#1}` is five characters

    def test_unsafe_1(self, bash):
        output = assert_bash_exec(
            bash, "! __tester '$(echo hello >&2)'", want_output=True
        )
        assert output.strip() == ""

    def test_unsafe_2(self, bash):
        output = assert_bash_exec(
            bash, "! __tester '|echo hello >&2'", want_output=True
        )
        assert output.strip() == ""

    def test_unsafe_3(self, bash):
        output = assert_bash_exec(
            bash, "! __tester '>| important_file.txt'", want_output=True
        )
        assert output.strip() == ""

    def test_unsafe_4(self, bash):
        output = assert_bash_exec(
            bash, "! __tester '`echo hello >&2`'", want_output=True
        )
        assert output.strip() == ""

    def test_glob_default(self, bash):
        with bash_env_saved(bash) as bash_env:
            bash_env.shopt("failglob", False)
            bash_env.shopt("nullglob", False)
            output = assert_bash_exec(
                bash, "__tester 'non-existent-*.txt'", want_output=True
            )
            assert output.strip() == "<non-existent-*.txt>"

    def test_glob_noglob(self, bash):
        with bash_env_saved(bash) as bash_env:
            bash_env.set("noglob", True)
            output = assert_bash_exec(
                bash,
                "__tester 'non-existent-*.txt'",
                want_output=True,
            )
            assert output.strip() == "<non-existent-*.txt>"

    def test_glob_failglob(self, bash):
        with bash_env_saved(bash) as bash_env:
            bash_env.shopt("failglob", True)
            output = assert_bash_exec(
                bash, "! __tester 'non-existent-*.txt'", want_output=True
            )
            assert output.strip() == ""

    def test_glob_nullglob(self, bash):
        with bash_env_saved(bash) as bash_env:
            bash_env.shopt("failglob", False)
            bash_env.shopt("nullglob", True)
            output = assert_bash_exec(
                bash, "__tester 'non-existent-*.txt'", want_output=True
            )
            assert output.strip() == ""



import pytest

from conftest import assert_complete


class TestUlimit:
    @pytest.mark.complete("ulimit ")
    def test_1(self, completion):
        assert not completion

    @pytest.mark.complete("ulimit -", require_cmd=True)
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("ulimit -S -")
    def test_3(self, completion):
        """Test modes are completed after -S (-S not treated as mode)."""
        assert completion

    @pytest.mark.complete("ulimit -u -")
    def test_4(self, completion):
        """Test modes are NOT completed if one is specified."""
        assert not completion

    @pytest.mark.complete("ulimit -c ")
    def test_5(self, completion):
        assert completion
        assert not any(x.startswith("-") for x in completion)

    @pytest.mark.complete("ulimit -a ")
    def test_6(self, completion):
        assert completion == sorted("-S -H".split())

    @pytest.mark.complete("ulimit -a -H -")
    def test_7(self, completion):
        """Test modes are NOT completed with -a given somewhere."""
        assert not completion

    @pytest.mark.parametrize("flag", ["-S", "-H"])
    def test_no_special_values_after_soft_or_hard(self, bash, flag):
        completion = assert_complete(bash, "ulimit %s " % flag)
        assert not completion

    @pytest.mark.complete("ulimit -c 0 -n ")
    def test_special_not_just_first(self, completion):
        """Test we offer limit values not only for the first option."""
        assert completion



import pytest

from conftest import assert_complete


class TestEnv:
    @pytest.mark.complete("env --", require_longopt=True)
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("env __unknown_variable__=")
    def test_unknown_variable_falls_back_to_filedir(self, completion):
        assert "shared/" in completion

    @pytest.mark.complete("env LANG=", xfail="! locale -a &>/dev/null")
    def test_lang_envvar(self, completion):
        assert any(x == "C" or x.startswith("C.") for x in completion)

    @pytest.mark.parametrize(
        "opts",
        [
            "",
            "foo=bar",
            "--debug",
            "--debug foo=bar",
            "-",
            "- foo=bar",
        ],
    )
    def test_command(self, bash, opts):
        completion = assert_complete(bash, "env %s s" % opts)
        assert completion == "h" or "sh" in completion

    @pytest.mark.parametrize(
        "opts",
        [
            "foo=bar --non-existent",
            "- --non-existent",
            "-- --non-existent",
        ],
    )
    def test_option_like_command(self, bash, opts):
        completion = assert_complete(bash, "env %s s" % opts)
        assert not (completion == "h" or "sh" in completion)



import pytest


class TestPushd:
    @pytest.mark.complete("pushd ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("pushd -")
    def test_options(self, completion):
        assert completion



import pytest


class TestMkdir:
    @pytest.mark.complete("mkdir ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("mkdir ", cwd="shared/default")
    def test_2(self, completion):
        assert completion == ["bar", "bar bar.d/", "foo", "foo.d/"]

    @pytest.mark.xfail  # TODO: why path in completion, basename in .output?
    @pytest.mark.complete("mkdir shared/default/foo.d/")
    def test_3(self, completion):
        assert completion.output == "foo"
        assert completion == [completion.output]

    @pytest.mark.complete("mkdir -", require_longopt=True)
    def test_options(self, completion):
        assert completion



import pytest


class TestInotifywatch:
    @pytest.mark.complete("inotifywatch ")
    def test_1(self, completion):
        assert completion

    @pytest.mark.complete("inotifywatch --", require_cmd=True)
    def test_2(self, completion):
        assert completion

    @pytest.mark.complete("inotifywatch -e ", require_cmd=True)
    def test_3(self, completion):
        assert len(completion) > 1



import re

import pytest

from conftest import assert_bash_exec


@pytest.mark.bashcomp(ignore_env=r"^-declare -f _comp_cmd_tar$")
class TestTar:
    @pytest.fixture(scope="class")
    def gnu_tar(self, bash):
        got = assert_bash_exec(bash, "tar --version || :", want_output=True)
        if not re.search(r"\bGNU ", got):
            pytest.skip("Not GNU tar")

    @pytest.mark.complete("tar ", shopt=dict(failglob=True))
    def test_1(self, bash, completion):
        assert completion

    # Test "f" when mode is not as first option
    @pytest.mark.complete("tar zfc ", cwd="tar")
    def test_2(self, completion):
        assert completion == "dir/ dir2/".split()

    @pytest.mark.complete("tar cf ", cwd="tar")
    def test_3(self, completion):
        assert completion == "dir/ dir2/".split()

    @pytest.mark.complete("tar tf archive.tar.xz dir/file", cwd="tar")
    def test_4(self, completion):
        assert completion == "dir/fileA dir/fileB dir/fileC".split()

    @pytest.mark.complete("tar cTfvv NOT_EXISTS DONT_CREATE.tar ", cwd="tar")
    def test_5(self, completion):
        assert completion == "archive.tar.xz dir/ dir2/ escape.tar".split()

    @pytest.mark.complete("tar xvf ", cwd="tar")
    def test_6(self, completion):
        assert completion == "archive.tar.xz dir/ dir2/ escape.tar".split()

    @pytest.mark.complete("tar -c")
    def test_7(self, completion, gnu_tar):
        """Test short options."""
        assert completion

    @pytest.mark.complete("tar -zcf ", cwd="tar")
    def test_8(self, completion, gnu_tar):
        """Test mode not as first option."""
        assert completion == "dir/ dir2/".split()

    @pytest.mark.complete("tar -cf ", cwd="tar")
    def test_9(self, completion, gnu_tar):
        """Test that we don't suggest rewriting existing archive."""
        assert completion == "dir/ dir2/".split()

    @pytest.mark.complete("tar -c --file ", cwd="tar")
    def test_10(self, completion, gnu_tar):
        assert completion == "dir/ dir2/".split()

    @pytest.mark.complete("tar -cvv --file ", cwd="tar")
    def test_11(self, completion, gnu_tar):
        assert completion == "dir/ dir2/".split()

    @pytest.mark.complete("tar -tf archive.tar.xz dir/file", cwd="tar")
    def test_12(self, completion, gnu_tar):
        """Test archive listing."""
        assert completion == "dir/fileA dir/fileB dir/fileC".split()

    @pytest.mark.complete("tar -t --file archive.tar.xz dir/file", cwd="tar")
    def test_13(self, completion, gnu_tar):
        """Test archive listing with --file."""
        assert completion == "dir/fileA dir/fileB dir/fileC".split()

    @pytest.mark.complete("tar --block")
    def test_14(self, completion, gnu_tar):
        assert completion == "--block-number --blocking-factor=".split()

    @pytest.mark.complete("tar --add-fil")
    def test_15(self, completion, gnu_tar):
        assert completion == "e="
        assert not completion.endswith(" ")

    @pytest.mark.complete("tar -cf /dev/null --posi")
    def test_16(self, completion, gnu_tar):
        assert completion == "x"
        assert completion.endswith(" ")

    @pytest.mark.complete("tar --owner=")
    def test_17(self, bash, completion, gnu_tar, output_sort_uniq):
        users = output_sort_uniq("compgen -u")
        assert completion == users

    @pytest.mark.complete("tar --group=")
    def test_18(self, bash, completion, gnu_tar, output_sort_uniq):
        groups = output_sort_uniq("compgen -g")
        assert completion == groups

    # Use -b for this as -b is still not handled by tar's completion
    @pytest.mark.complete("tar -cvvfb ")
    def test_19(self, bash, completion, gnu_tar):
        """Test short option -XXXb <TAB> (arg required)."""
        assert not completion

    # Use bsdtar here as it completes to only 'zc zt zx'
    # -- 'tar' can be GNU tar and have more options
    @pytest.mark.complete("bsdtar z")
    def test_20(self, bash, completion):
        assert completion == "zc zt zx".split()

    @pytest.mark.complete("bsdtar cbfvv NON_EXISTENT ", cwd="tar")
    def test_21(self, bash, completion):
        """Test _second_ option in "old" argument."""
        assert completion == "dir/ dir2/".split()

    @pytest.mark.complete(r"tar tf escape.tar a/b\'", cwd="tar")
    def test_22(self, bash, completion):
        """Test listing escaped chars in old option."""
        assert completion == "c/"

    # Test long options (gnu style)
    @pytest.mark.complete("tar --create --file created.tar ", cwd="tar")
    def test_23(self, completion):
        assert completion == "archive.tar.xz dir/ dir2/ escape.tar".split()

    # Test detection of file type by extension, J is for .xz
    @pytest.mark.complete("tar xJf ", cwd="tar")
    def test_24(self, completion):
        assert completion == "archive.tar.xz dir/ dir2/".split()

    # Test compression detection of gnu style options
    @pytest.mark.complete("tar --extract --xz --file ", cwd="tar")
    def test_25(self, completion):
        assert completion == "archive.tar.xz dir/ dir2/".split()

    # TODO: "tar tf escape.tar a/b"



import os.path

import pytest


@pytest.mark.bashcomp(cmd="dpkg-query")
class TestDpkgQuery:
    @pytest.mark.complete("dpkg-query --", require_cmd=True)
    def test_options(self, completion):
        assert completion

    @pytest.mark.xfail(
        not os.path.exists("/etc/debian_version"),
        reason="Likely fails on systems not based on Debian",
    )
    @pytest.mark.complete(
        "dpkg-query -W dpk",
        require_cmd=True,
        xfail="! apt-cache show &>/dev/null",  # empty cache?
    )
    def test_show(self, completion):
        assert "dpkg" in completion



import pytest

from conftest import assert_bash_exec, bash_env_saved


@pytest.mark.bashcomp(
    cmd=None, ignore_env=r"^\+declare -f (dump_array|__tester)$"
)
class TestUtilSplit:
    @pytest.fixture
    def functions(self, bash):
        assert_bash_exec(
            bash, "dump_array() { printf '<%s>' \"${arr[@]}\"; echo; }"
        )
        assert_bash_exec(
            bash,
            '__tester() { local -a arr=(00); _comp_split "${@:1:$#-1}" arr "${@:$#}"; dump_array; }',
        )

    def test_1(self, bash, functions):
        output = assert_bash_exec(
            bash, "__tester '12 34 56'", want_output=True
        )
        assert output.strip() == "<12><34><56>"

    def test_2(self, bash, functions):
        output = assert_bash_exec(
            bash, "__tester $'12\\n34\\n56'", want_output=True
        )
        assert output.strip() == "<12><34><56>"

    def test_3(self, bash, functions):
        output = assert_bash_exec(
            bash, "__tester '12:34:56'", want_output=True
        )
        assert output.strip() == "<12:34:56>"

    def test_option_F_1(self, bash, functions):
        output = assert_bash_exec(
            bash, "__tester -F : '12:34:56'", want_output=True
        )
        assert output.strip() == "<12><34><56>"

    def test_option_F_2(self, bash, functions):
        output = assert_bash_exec(
            bash, "__tester -F : '12 34 56'", want_output=True
        )
        assert output.strip() == "<12 34 56>"

    def test_option_l_1(self, bash, functions):
        output = assert_bash_exec(
            bash, "__tester -l $'12\\n34\\n56'", want_output=True
        )
        assert output.strip() == "<12><34><56>"

    def test_option_l_2(self, bash, functions):
        output = assert_bash_exec(
            bash, "__tester -l '12 34 56'", want_output=True
        )
        assert output.strip() == "<12 34 56>"

    def test_option_a_1(self, bash, functions):
        output = assert_bash_exec(
            bash, "__tester -aF : '12:34:56'", want_output=True
        )
        assert output.strip() == "<00><12><34><56>"

    def test_protect_from_failglob(self, bash, functions):
        with bash_env_saved(bash) as bash_env:
            bash_env.shopt("failglob", True)
            output = assert_bash_exec(
                bash, "__tester -F '*' '12*34*56'", want_output=True
            )
            assert output.strip() == "<12><34><56>"

    def test_protect_from_nullglob(self, bash, functions):
        with bash_env_saved(bash) as bash_env:
            bash_env.shopt("nullglob", True)
            output = assert_bash_exec(
                bash, "__tester -F '*' '12*34*56'", want_output=True
            )
            assert output.strip() == "<12><34><56>"

    def test_protect_from_IFS(self, bash, functions):
        with bash_env_saved(bash) as bash_env:
            bash_env.write_variable("IFS", "34")
            output = assert_bash_exec(
                bash, "__tester '12 34 56'", want_output=True
            )
            assert output.strip() == "<12><34><56>"



import pytest

from conftest import assert_bash_exec


@pytest.mark.bashcomp(cmd=None)
class TestUnitExpandTilde:
    def test_1(self, bash):
        """The old interface `__expand_tilde_by_ref` should not fail when it is
        called without arguments"""
        assert_bash_exec(bash, "__expand_tilde_by_ref >/dev/null")

    def test_2(self, bash):
        """Test environment non-pollution, detected at teardown."""
        assert_bash_exec(
            bash,
            '_x() { local REPLY; _comp_expand_tilde "~"; }; _x; unset -f _x',
        )

    @pytest.fixture(scope="class")
    def functions(self, bash):
        # $HOME tinkering: protect against $HOME != ~user; our "home" is the
        # latter but plain_tilde follows $HOME
        assert_bash_exec(
            bash,
            '_comp__test_unit() { local REPLY HOME=$1; _comp_expand_tilde "$2"; printf "%s\\n" "$REPLY"; }',
        )

    @pytest.mark.parametrize("plain_tilde", (True, False))
    @pytest.mark.parametrize(
        "suffix_expanded",
        (
            ("", True),
            ("/foo", True),
            (r"/\$HOME", True),
            ("/a  b", True),
            ("/*", True),
            (";echo hello", False),
            ("/a;echo hello", True),
        ),
    )
    def test_expand(
        self, bash, user_home, plain_tilde, suffix_expanded, functions
    ):
        user, home = user_home
        suffix, expanded = suffix_expanded
        home2 = home
        if plain_tilde:
            user = ""
            if not suffix or not expanded:
                home2 = "~"
        elif not expanded:
            home2 = "~%s" % user
        output = assert_bash_exec(
            bash,
            r'_comp__test_unit "%s" "~%s%s"' % (home, user, suffix),
            want_output=True,
        )
        assert output.strip() == "%s%s" % (home2, suffix.replace(r"\$", "$"))
